from collections import OrderedDict
from functools import reduce
import itertools
import json
import math
import os.path
import sys
from typing import Dict, Sequence, Union, Tuple, Optional, List
import warnings

import numpy as np
import seekpath
from seekpath.hpkot import SymmetryDetectionError
from importlib_resources import files
from pint import UndefinedUnitError

from euphonic import ureg, Quantity
import euphonic.data


def direction_changed(qpts: np.ndarray, tolerance: float = 5e-6
                      ) -> np.ndarray:
    """
    Determines whether the q-direction has changed between each pair of
    q-points

    Parameters
    ----------
    qpts
        Shape (n_qpts, 3) float ndarray. The q-points to use
    tolerance
        The minimum difference between the dot product and magnitude of
        both vectors multiplied together for each pair of direction
        vectors required for the direction to have changed

    Returns
    -------
    has_direction_changed
        Shape (n_qpts - 2,) bool ndarray. Indicates whether the
        direction vector has changed between each pair of q-points
    """

    # Get vectors between q-points
    delta = np.diff(qpts, axis=0)

    # Dot each vector with the next to determine the relative direction
    dot = np.einsum('ij,ij->i', delta[1:, :], delta[:-1, :])
    # Get magnitude of each vector
    modq = np.linalg.norm(delta, axis=1)
    # Determine how much the direction has changed (dot) relative to the
    # vector sizes (modq)
    direction_changed = (np.abs(np.abs(dot) - modq[1:]*modq[:-1]) > tolerance)

    return direction_changed


def is_gamma(qpt: np.ndarray) -> Union[bool, np.ndarray]:
    """
    Determines whether the given point(s) are gamma points

    Parameters
    ----------
    qpts
        Shape (3,) or (N, 3) float ndarray. The q-point or q-points

    Returns
    -------
    isgamma
        bool or shape (N,) bool ndarray. Whether the input q-points(s)
        are gamma points. Returns a scalar if only 1 q-point is
        provided
    """
    tol = 1e-15
    isgamma = np.sum(np.absolute(qpt - np.rint(qpt)), axis=-1) < tol
    return isgamma


def mp_grid(grid: Tuple[int, int, int]) -> np.ndarray:
    """
    Returns the q-points on a MxNxL Monkhorst-Pack grid specified by
    grid

    Parameters
    ----------
    grid
        The number of points in each direction

    Returns
    -------
    qgrid
        Shape (M*N*L, 3) float ndarray. Q-points on an MP grid
    """
    # Monkhorst-Pack grid: ur = (2r-qr-1)/2qr where r=1,2..,qr
    qh = np.true_divide(
        2*(np.arange(grid[0]) + 1) - grid[0] - 1, 2*grid[0])
    qh = np.repeat(qh, grid[1]*grid[2])
    qk = np.true_divide(
        2*(np.arange(grid[1]) + 1) - grid[1] - 1, 2*grid[1])
    qk = np.tile(np.repeat(qk, grid[2]), grid[0])
    ql = np.true_divide(
        2*(np.arange(grid[2]) + 1) - grid[2] - 1, 2*grid[2])
    ql = np.tile(ql, grid[0]*grid[1])
    return np.column_stack((qh, qk, ql))


def get_all_origins(max_xyz: Tuple[int, int, int],
                    min_xyz: Tuple[int, int, int] = (0, 0, 0),
                    step: int = 1) -> np.ndarray:
    """
    Given the max/min number of cells in each direction, get a list of
    all possible cell origins

    Parameters
    ----------
    max_xyz
        The number of cells to count to in each direction
    min_xyz
        The cell number to count from in each direction
    step
        The step between cells

    Returns
    -------
    origins
        Shape (prod(max_xyz - min_xyz)/step, 3) int ndarray. The cell
        origins
    """
    diff = np.absolute(np.subtract(max_xyz, min_xyz))
    nx = np.repeat(range(min_xyz[0], max_xyz[0], step), diff[1]*diff[2])
    ny = np.repeat(np.tile(range(min_xyz[1], max_xyz[1], step), diff[0]),
                   diff[2])
    nz = np.tile(range(min_xyz[2], max_xyz[2], step), diff[0]*diff[1])

    return np.column_stack((nx, ny, nz))


def get_qpoint_labels(qpts: np.ndarray,
                      cell: Optional[Tuple[List[List[float]],
                                           List[List[float]],
                                           List[int]]] = None
                      ) -> List[Tuple[int, str]]:
    """
    Gets q-point labels (e.g. GAMMA, X, L) for the q-points at which the
    path through reciprocal space changes direction, or where a point
    appears twice in succession.

    Parameters
    ----------
    qpts
        Shape (n_qpts, 3) float ndarray. The q-points to get labels for
    cell
        The cell structure as defined by spglib. Can be obtained by
        Crystal.to_spglib_cell. If not provided, the labels will be
        generic e.g. '1/3 1/2 0' rather than high-symmetry point labels

    Returns
    -------
    x_tick_labels
        Tick labels and the q-point indices that they apply to
    """
    xlabels, qpts_with_labels = _recip_space_labels(qpts, cell=cell)
    for i, label in enumerate(xlabels):
        if label == 'GAMMA':
            xlabels[i] = r'$\Gamma$'
    if np.all(xlabels == ''):
        xlabels = [str(x) for x in np.around(
            qpts[qpts_with_labels, :], decimals=2)]
    qpts_with_labels = [int(x) for x in qpts_with_labels.tolist()]
    return list(zip(qpts_with_labels, xlabels))


def get_reference_data(collection: str = 'Sears1992',
                       physical_property: str = 'coherent_scattering_length'
                       ) -> Dict[str, Quantity]:
    """
    Get physical data as a dict of (possibly-complex) floats from reference
    data.

    Each "collection" refers to a JSON file which may contain any set of
    properties, indexed by physical_property.

    Properties are stored in JSON files, encoding a single dictionary with the
    structure::

      {"metadata1": "metadata1 text", "metadata2": ...,
       "physical_properties": {"property1": {"__units__": "unit_str",
                                             "H": H_property1_value,
                                             "He": He_property1_value,
                                             "Li": {"__complex__": true,
                                                    "real": Li_property1_real,
                                                    "imag": Li_property1_imag},
                                             "Nh": None,
                                             ...},
                               "property2": ...}}

    Parameters
    ----------
    collection
        Identifier of data file; this may be an inbuilt data set ("Sears1992"
        or "BlueBook") or a path to a JSON file (e.g. "./my_custom_data.json").

    physical_property
        The name of the property for which data should be extracted. This must
        match an entry of "physical_properties" in the data file.

    Returns
    -------
    Dict[str, Quantity]
        Requested data as a dict with string keys and (possibly-complex)
        float Quantity values. String or None items of the original data file
        will be omitted.

    """

    _reference_data_files = {'Sears1992': 'sears-1992.json',
                             'BlueBook': 'bluebook.json'}

    def custom_decode(dct):
        if '__complex__' in dct:
            return complex(dct['real'], dct['imag'])
        return dct

    if collection in _reference_data_files:
        filename = _reference_data_files[collection]
        with open(files(euphonic.data) / filename) as fd:
            file_data = json.load(fd, object_hook=custom_decode)

    elif os.path.isfile(collection):
        filename = collection
        with open(filename) as fd:
            file_data = json.load(fd, object_hook=custom_decode)
    else:
        raise ValueError(
            f'No data files known for collection "{collection}". '
            f'Available collections: '
            + ', '.join(list(_reference_data_files)))

    if 'physical_property' not in file_data:
        raise AttributeError('Data file does not contain required key '
                             '"physical_property".')

    data = file_data['physical_property'].get(physical_property)
    if data is None:
        raise ValueError(
            f'No such collection "{collection}" with property '
            f'"{physical_property}". Available properties for this collection'
            ': ' + ', '.join(list(file_data["physical_property"].keys())))

    unit_str = data.get('__units__')
    if unit_str is None:
        raise ValueError(f'Reference data file "{filename}" does not '
                         'specify dimensions with "__units__" metadata.')

    try:
        unit = ureg(unit_str)
    except UndefinedUnitError:
        raise ValueError(
            f'Units "{unit_str}" from data file "{filename}" '
            'are not supported by the Euphonic unit register.')

    return {key: value * unit
            for key, value in data.items()
            if isinstance(value, (float, complex))}


def mode_gradients_to_widths(mode_gradients: Quantity, cell_vectors: Quantity
                             ) -> Quantity:
    """
    Converts either scalar or vector mode gradients (units
    energy/(length^-1)) to an estimate of the mode widths (units
    energy). If vector mode gradients are supplied, they will first be
    converted to scalar gradients by taking the Frobenius norm (using
    np.linalg.norm). The conversion to mode widths uses the cell volume
    and number of q-points to estimate the q-spacing. Note that the
    number of q-points is determined by the size of mode_gradients, so
    is not likely to give accurate widths if the q-points have been
    symmetry reduced.

    Parameters
    ----------
    mode_gradients
        Shape (n_qpts, n_modes) float Quantity if using scalar mode
        gradients, shape (n_qpts, n_modes, 3) float Quantity if using
        vector mode gradients
    cell_vectors
        Shape (3, 3) float Quantity. The cell vectors
    """
    modg = mode_gradients.to('hartree*bohr').magnitude
    if modg.ndim == 3 and modg.shape[2] == 3:
        modg = np.linalg.norm(modg, axis=2)
    elif modg.ndim != 2:
        raise ValueError(
            f'Unexpected shape for mode_gradients {modg.shape}, '
            f'expected (n_qpts, n_modes) or (n_qpts, n_modes, 3)')
    cell_volume = (_cell_vectors_to_volume(
        cell_vectors.magnitude)*cell_vectors.units**3).to('bohr**3').magnitude
    q_spacing = 2/(np.cbrt(len(mode_gradients)*cell_volume))
    mode_widths = q_spacing*modg
    return mode_widths*ureg('hartree').to(
        mode_gradients.units/cell_vectors.units)


def convert_fc_phases(force_constants: np.ndarray, atom_r: np.ndarray,
                      sc_atom_r: np.ndarray, uc_to_sc_atom_idx: np.ndarray,
                      sc_to_uc_atom_idx: np.ndarray, sc_matrix: np.ndarray,
                      cell_origins_tol: float = 1e-5
                      ) -> Tuple[np.ndarray, np.ndarray]:
    """
    Convert from a force constants matrix which uses the atom
    coordinates as r in the e^-iq.r phase (Phonopy-like), to a
    matrix which uses the cell origin coordinates as r in the
    e^-iq.r phase (Euphonic, CASTEP-like). Also changes the shape
    of the matrix to be compatible with Euphonic.

    Parameters
    ----------
    force_constants
        Shape (n_atoms, n_cells*n_atoms, 3, 3) or
        (n_cells*n_atoms, n_cells*n_atoms, 3, 3) float ndarray with
        phases based on the atom coordinate. n_atoms is the number
        of atoms in the unit cell, n_cells is the number of unit cells
        in the supercell
    atom_r
        Shape (n_atoms, 3) float ndarray. The coordinates of the atoms
        in the unit cell in fractional coordinates of the unit cell
        vectors
    sc_atom_r
        Shape (n_atoms_in_sc, 3) float ndarray. The coordinates of the
        atoms in the supercell in fractional coordinates of the unit
        cell vectors. n_atoms_in_sc is the number of atoms in the
        supercell
    uc_to_sc_atom_idx
        Shape (n_atoms,) int ndarray. For each atom in the unit cell,
        the index of that atom in the supercell
    sc_to_uc_atom_idx
        Shape (n_atoms_in_sc,) int ndarray. For each atom in the
        supercell, the index of the equivalent atom in
        the unit cell
    sc_matrix
        Shape (3, 3) int ndarray. The transformation matrix to convert
        from the unit cell vectors to the supercell vectors
    cell_origins_tol
        The calculated cell origins for each atom should be integer
        values. This defines the tolerance as above which a value is
        considered non-integer, which may indicate an issue with the
        input and will raise a RuntimeError.

    Returns
    -------
    fc_converted
        Shape (n_cells, 3*n_atoms, 3*n_atoms) float ndarray. The force
        constants matrix with a phase and shape compatible with
        Euphonic
    cell_origins
        Shape (n_cells_in_sc, 3) int ndarray. The origin coordinates of
        each unit cell within the supercell, in units of the unit cell
        vectors
    """
    n_atoms_sc = len(sc_to_uc_atom_idx)
    n_atoms_uc = len(uc_to_sc_atom_idx)
    n_cells = int(np.rint(np.absolute(np.linalg.det(sc_matrix))))
    if n_atoms_sc/n_atoms_uc - n_cells != 0:
        raise ValueError(
            f'Inconsistent numbers of cells in the supercell, unit '
            f'cell has {n_atoms_uc} atoms, and supercell has '
            f'{n_atoms_sc} atoms, but sc_matrix determinant suggests '
            f'there should be {n_cells} cells in the supercell')
    # Get cell origins for all atoms
    cell_origins_per_atom = sc_atom_r - atom_r[sc_to_uc_atom_idx]
    non_int = np.where(
        np.abs(cell_origins_per_atom
               - np.rint(cell_origins_per_atom)) > cell_origins_tol)[0]
    if len(non_int) > 0:
        raise RuntimeError(
            f'Non-integer cell origins for atom(s) '
            f'{", ".join(np.unique(non_int).astype(str))}, '
            f'check coordinates and indices are correct')
    cell_origins_per_atom = np.rint(cell_origins_per_atom).astype(np.int32)
    # Recenter cell origins onto those for atom 0 in unit cell 0
    cell_origins_per_atom -= cell_origins_per_atom[uc_to_sc_atom_idx[0]]

    # Build unique cell origins by getting  cell origins
    # associated with primitive atom 0 in all unit cells
    atom0_idx = np.where(sc_to_uc_atom_idx == 0)[0]
    cell_origins = cell_origins_per_atom[atom0_idx]
    # For some supercells, cell origins aren't always the
    # same for each atom in a supercell, and the cell origins are
    # sometimes outside the supercell. Create a mapping of cell
    # origins for atoms 1..n onto the equivalent cell origins for
    # atom 0, so the same cell origins can be used for all atoms
    cell_origins_map = np.zeros((n_atoms_sc), dtype=np.int32)
    # Get origins of adjacent supercells in prim cell frac coords
    sc_origins =  get_all_origins((2,2,2), min_xyz=(-1,-1,-1))
    sc_origins_pcell = np.einsum('ij,jk->ik', sc_origins, sc_matrix)
    for i in range(n_atoms_sc):
        co_idx = np.where(
            (cell_origins_per_atom[i] == cell_origins).all(axis=1))[0]
        if len(co_idx) != 1:
            # Get equivalent cell origin in surrounding supercells
            origin_in_scs = cell_origins_per_atom[i] - sc_origins_pcell
            co_idx = -1
            # Find which of the 'unique' cell origins is equivalent
            for j, cell_origin in enumerate(cell_origins):
                if np.any((origin_in_scs == cell_origin).all(axis=1)):
                    co_idx = j
                    break
            if co_idx == -1:
                raise Exception((
                    'Couldn\'t determine cell origins for '
                    'force constants matrix'))
        cell_origins_map[i] = co_idx

    if force_constants.shape[0] == force_constants.shape[1]:
        # full fc
        force_constants = force_constants[uc_to_sc_atom_idx]
    fc_converted = np.full((n_atoms_uc, n_cells, n_atoms_uc, 3, 3), -1.0)
    for i in range(n_atoms_uc):
        fc_tmp = np.zeros((n_cells, n_atoms_uc, 3, 3))
        fc_tmp[cell_origins_map, sc_to_uc_atom_idx] = force_constants[i]
        # For Phonopy force constants, the n_atoms in the (n_atoms, ...)
        # shaped array may not be in the same cell within the supercell,
        # but Euphonic's interpolation requires this, so use equivalent
        # cell vector indices to arrange the force constants correctly
        atom_idx = np.where(sc_to_uc_atom_idx == i)[0][0]
        cell_idx = cell_origins_map[atom_idx]
        sc_relative_idx = _get_supercell_relative_idx(cell_origins, sc_matrix)
        fc_converted[i, sc_relative_idx[cell_idx]] = fc_tmp

    fc_converted =  np.reshape(np.transpose(
        fc_converted,
        axes=[1, 0, 3, 2, 4]), (n_cells, 3*n_atoms_uc, 3*n_atoms_uc))
    return fc_converted, cell_origins


def _cell_vectors_to_volume(cell_vectors: np.ndarray) -> float:
    """Convert 3x3 cell vectors to volume"""
    return np.dot(cell_vectors[0],
                  np.cross(cell_vectors[1], cell_vectors[2]))


def _get_unique_elems_and_idx(
        all_elems: Sequence[Tuple[Union[int, str], ...]]
        ) -> 'OrderedDict[Tuple[Union[int, str], ...], np.ndarray]':
    """
    Returns an ordered dictionary mapping the unique sequences of
    elements to their indices
    """
    # Abuse OrderedDict to get ordered set
    unique_elems = OrderedDict(
        zip(all_elems, itertools.cycle([None]))).keys()
    return OrderedDict((
        elem,
        np.asarray([i for i, other_elem in enumerate(all_elems)
                    if elem == other_elem])
        ) for elem in unique_elems)


def _calc_abscissa(reciprocal_cell: Quantity, qpts: np.ndarray
                       ) -> Quantity:
    """
    Calculates the distance between q-points (e.g. to use as a plot
    x-coordinate)

    Parameters
    ----------
    reciprocal_cell
        Shape (3, 3) float Quantity. The reciprocal cell, can be
        calculated with Crystal.reciprocal_cell
    qpts
        Shape (n_qpts, 3) float ndarray. The q-points to get the
        distance between, in reciprocal lattice units

    Returns
    -------
    abscissa
        Shape (n_qpts) float Quantity. The distance between q-points
        in 1/crystal.cell_vectors_unit
    """
    recip = reciprocal_cell.to('1/bohr').magnitude
    # Get distance between q-points in each dimension
    # Note: length is nqpts - 1
    delta = np.diff(qpts, axis=0)

    # Determine how close delta is to being an integer. As q = q + G
    # where G is a reciprocal lattice vector based on {1,0,0},{0,1,0},
    # {0,0,1}, this is used to determine whether 2 q-points are
    # equivalent ie. they differ by a multiple of the reciprocal lattice
    # vector
    delta_rem = np.sum(np.abs(delta - np.rint(delta)), axis=1)

    # Create a boolean array that determines whether to calculate the
    # distance between q-points,taking into account q-point equivalence.
    # If delta is more than the tolerance, but delta_rem is less than
    # the tolerance, the q-points differ by G so are equivalent and the
    # distance shouldn't be calculated
    TOL = 0.001
    calc_modq = np.logical_not(np.logical_and(
        np.sum(np.abs(delta), axis=1) > TOL,
        delta_rem < TOL))

    # Multiply each delta by the recip lattice to get delta in Cartesian
    deltaq = np.einsum('ji,kj->ki', recip, delta)

    # Get distance between q-points for all valid pairs of q-points
    modq = np.zeros(np.size(delta, axis=0))
    modq[calc_modq] = np.sqrt(np.sum(np.square(deltaq[calc_modq]), axis=1))

    # Prepend initial x axis value of 0
    abscissa = np.insert(modq, 0, 0.)

    # Do cumulative sum to get position along x axis
    abscissa = np.cumsum(abscissa)
    return abscissa*ureg('1/bohr').to(reciprocal_cell.units)


def _recip_space_labels(qpts: np.ndarray,
                        cell: Optional[Tuple[List[List[float]],
                                             List[List[float]],
                                             List[int]]]
                        ) -> Tuple[np.ndarray, np.ndarray]:
    """
    Gets q-points point labels (e.g. GAMMA, X, L) for the q-points at
    which the path through reciprocal space changes direction or where a
    q-point is repeated

    Parameters
    ----------
    qpts
        Shape (n_qpts, 3) float ndarray. The q-points to get labels for
    cell
        The cell structure as defined by spglib. Can be obtained by
        Crystal.to_spglib_cell. If not provided, the labels will be
        generic e.g. '1/3 1/2 0' rather than high-symmetry point labels

    Returns
    -------
    labels
        Shape (n_qpts_direction_changed,) string ndarray. The labels
        for each q-point at which the path through reciprocal space
        changes direction
    qpts_with_labels
        Shape (n_qpts_direction_changed,) int ndarray. The indices of
        the q-points at which the path through reciprocal space changes
        direction
    """

    # First and last q-points should always be labelled
    if len(qpts) <= 2:
        qpt_has_label = np.ones(len(qpts), dtype=bool)
    else:
        qpt_has_label = np.concatenate((
            [True],
            reduce(np.logical_or,
                   (direction_changed(qpts),
                    is_gamma(qpts[1:-1]),
                    np.all(qpts[2:] == qpts[1:-1], axis=1)
                    )
                   ),
            [True]))
    qpts_with_labels = np.where(qpt_has_label)[0]

    if cell is None:
        sym_label_to_coords = _generic_qpt_labels()
    else:
        try:
            sym_label_to_coords = seekpath.get_path(cell)["point_coords"]
        except SymmetryDetectionError:
            warnings.warn(('Could not determine cell symmetry, using generic '
                           'q-point labels'), stacklevel=2)
            sym_label_to_coords = _generic_qpt_labels()

    # Get labels for each q-point
    labels = np.array([])
    for qpt in qpts[qpts_with_labels]:
        labels = np.append(labels, _get_qpt_label(qpt, sym_label_to_coords))

    return labels, qpts_with_labels


def _generic_qpt_labels() -> Dict[str, Tuple[float, float, float]]:
    """
    Returns a dictionary relating fractional q-point label strings to
    their coordinates e.g. '1/4 1/2 1/4' = [0.25, 0.5, 0.25]. Used for
    labelling q-points when the space group can't be calculated
    """
    label_strings = ['0', '1/4', '3/4', '1/2', '1/3', '2/3', '3/8', '5/8']
    label_coords = [0., 0.25, 0.75, 0.5, 1./3., 2./3., 0.375, 0.625]

    generic_labels = {}
    for i, s1 in enumerate(label_strings):
        for j, s2 in enumerate(label_strings):
            for k, s3 in enumerate(label_strings):
                key = s1 + ' ' + s2 + ' ' + s3
                value = (label_coords[i], label_coords[j], label_coords[k])
                generic_labels[key] = value
    return generic_labels


def _get_qpt_label(qpt: np.ndarray,
                   point_labels: Dict[str, Tuple[float, float, float]]
                   ) -> str:
    """
    Gets a label for a particular q-point, based on the high symmetry
    points of a particular space group. Used for labelling the
    dispersion plot x-axis

    Parameters
    ----------
    qpt
        Shape (3,) float ndarray. Single q-point coordinates
    point_labels
        A dictionary with N entries, relating high symmetry point labels
        (e.g. 'GAMMA', 'X'), to their 3-dimensional coordinates (e.g.
        [0.0, 0.0, 0.0]) where N = number of high symmetry points for a
        particular space group

    Returns
    -------
    label
        The label for this q-point. If the q-point isn't a high symmetry
        point label is just an empty string
    """

    # Normalise qpt to [0,1]
    qpt_norm = [x - math.floor(x) for x in qpt]

    # Split dict into keys and values so labels can be looked up by
    # comparing q-point coordinates with the dict values
    labels = list(point_labels)
    # Ensure symmetry points in label_keys and label_values are in the
    # same order (not guaranteed if using .values() function)
    label_coords = [point_labels[x] for x in labels]

    # Check for matching symmetry point coordinates (roll q-point
    # coordinates if no match is found)
    TOL = 1e-6
    matching_label_index = np.where((np.isclose(
        label_coords, qpt_norm, atol=TOL)).all(axis=1))[0]
    if matching_label_index.size == 0:
        matching_label_index = np.where((np.isclose(
            label_coords, np.roll(qpt_norm, 1), atol=TOL)).all(axis=1))[0]
    if matching_label_index.size == 0:
        matching_label_index = np.where((np.isclose(
            label_coords, np.roll(qpt_norm, 2), atol=TOL)).all(axis=1))[0]

    label = ''
    if matching_label_index.size > 0:
        label = labels[matching_label_index[0]]

    return label


def _get_supercell_relative_idx(cell_origins: np.ndarray,
                                sc_matrix: np.ndarray) -> np.ndarray:
    """"
    For each cell_origins[i] -> cell_origins[j] vector in the supercell,
    gets the index n of the equivalent cell_origins[n] vector, where
    supercell_relative_idx[i, j] = n. Is required for converting from a
    compact to full force constants matrix

    Parameters
    ----------
    cell_origins
        Shape (n_cells, 3) int ndarray. The vector to the origin of
        each cell in the supercell, in unit cell fractional coordinates
    sc_matrix
        Shape (3, 3) int ndarray. The matrix for converting from the
        unit cell to the supercell

    Returns
    -------
    supercell_relative_idx
        Shape (n_cells, n_cells) int ndarray. The index n of the
        equivalent vector in cell_origins
    """
    n_cells = len(cell_origins)
    ax = np.newaxis

    # Get cell origins in supercell fractional coordinates
    inv_sc_matrix = np.linalg.inv(np.transpose(sc_matrix))
    cell_origins_sc = np.einsum('ij,kj->ik', cell_origins, inv_sc_matrix)
    sc_relative_index = np.zeros((n_cells, n_cells), dtype=np.int32)
    for nc in range(n_cells):
        # Get vectors from cell origin for a particular cell to all
        # other cell origins
        inter_cell_vectors = (cell_origins_sc
                              - np.tile(cell_origins_sc[nc], (n_cells, 1)))
        # Compare cell-cell vectors with vectors from cell 0's cell
        # origin to all other cell origins and determine which are
        # equivalent
        # Do calculation in chunks, so loop can be broken if all
        # equivalent vectors have been found
        N = 100
        dist_min = np.full((n_cells), sys.float_info.max)
        for i in range(int((n_cells - 1)/N) + 1):
            ci = i*N
            cf = min((i + 1)*N, n_cells)
            dist = (inter_cell_vectors[:, ax, :]
                    - cell_origins_sc[ax, ci:cf, :])
            dist_frac = dist - np.rint(dist)
            dist_frac_sum = np.sum(np.abs(dist_frac), axis=2)
            scri_current = np.argmin(dist_frac_sum, axis=1)
            dist_min_current = dist_frac_sum[
                range(n_cells), scri_current]
            replace = dist_min_current < dist_min
            sc_relative_index[nc, replace] = ci + scri_current[replace]
            dist_min[replace] = dist_min_current[replace]
            if np.all(dist_min <= 16*sys.float_info.epsilon):
                break
        if np.any(dist_min > 16*sys.float_info.epsilon):
            raise Exception('Couldn\'t find supercell relative index')
    return sc_relative_index


def _deprecation_warn(old_arg: str, new_arg: str, stacklevel: int = 3):
    """
    Emits a deprecation warning with a generic message
    """
    warnings.warn(f'{old_arg} has been deprecated and will be removed '
                  f'in a future release. Please use {new_arg} instead.',
                  category=DeprecationWarning,
                  stacklevel=stacklevel)



# -*- coding: UTF-8 -*-
"""
Run and time the Ewald sum part of an interpolation calculation for a
small number of q-points for different values of dipole_parameter (this
is the balance between the real/reciprocal space cutoffs when
performing the Ewald sum for polar materials) to determine the optimal
value
"""

from argparse import ArgumentParser
import time
from typing import List, Tuple, Optional
import warnings

import numpy as np

from euphonic import ForceConstants
from euphonic.cli.utils import (_get_cli_parser, get_args,
                                load_data_from_file)


def main(params: Optional[List[str]] = None) -> None:
    args = get_args(get_parser(), params)
    params = vars(args)
    params.update({'print_to_terminal': True})
    calculate_optimum_dipole_parameter(**params)


def calculate_optimum_dipole_parameter(
        filename: str,
        dipole_parameter_min: float = 0.25,
        dipole_parameter_max: float = 1.5,
        dipole_parameter_step: float = 0.25,
        n: int = 500,
        print_to_terminal: bool = False,
        **calc_modes_kwargs
        ) -> Tuple[float, float, float, np.ndarray, np.ndarray, np.ndarray]:
    """
    Calculate the optimum dipole_parameter and other dipole_parameters
    from the filename castep_bin file

    Parameters
    ----------
    filename : str
        The filename to read from
    dipole_parameter_min
        The minimum value of dipole_parameter to test
    dipole_parameter_max
        The maximum value of dipole_parameter to test
    dipole_parameter_step
        The difference between each dipole_parameter to test
    n
        The number of times to loop over q-points. A higher value will
        get a more reliable timing, but will take longer
    print_to_terminal
        Whether to print the outcome to terminal or not
    **calc_modes_kwargs
        Will be passed through to calculate_qpoint_phonon_modes

    Returns
    -------
    results
        A tuple of:
         - the optimal dipole_parameter
         - the time it took to initialise the optimal dipole_parameter,
         - the time per qpoint for the optimal dipole_parameter
         - other tested dipole_parameters
         - other tested diple_parameters initialisation times
         - other tested diple_parameters times per q-point
    """
    dipole_parameters = np.arange(
        dipole_parameter_min,
        dipole_parameter_max + dipole_parameter_step / 100,
        dipole_parameter_step)
    t_init = np.zeros(len(dipole_parameters), dtype=np.float64)
    t_per_qpt = np.zeros(len(dipole_parameters), dtype=np.float64)

    fc = load_data_from_file(filename)
    if not isinstance(fc, ForceConstants):
        raise TypeError('Force constants are required to use the '
                        'euphonic-optimise-dipole-parameter tool')
    # Only warn rather than error - although not designed for it
    # this script could still be useful for getting approximate
    # per-qpt timings for any material
    if fc.born is None:
        warnings.warn('Born charges not found for this material - '
                      'changing dipole_parameter will have no effect.')
    sfmt = '{:20s}'
    tfmt = '{: 3.2f}'
    dparamfmt = '{: 2.2f}'
    for i, dipole_parameter in enumerate(dipole_parameters):
        if print_to_terminal:
            print(('Results for dipole_parameter ' + dparamfmt).format(
                dipole_parameter))

        # Time Ewald sum initialisation
        start = time.time()
        fc.calculate_qpoint_phonon_modes(
            np.full((1, 3), 0.5), dipole_parameter=dipole_parameter,
            **calc_modes_kwargs)
        end = time.time()
        t_init[i] = end - start
        if print_to_terminal:
            print((sfmt + ': ' + tfmt + ' s').format(
                'Initialisation Time', t_init[i]))

        # Time per qpt
        qpts = np.full((n, 3), 0.5)
        start = time.time()
        # Need reduce_qpts=False because all q-points are the same,
        # if reduce_qpts=True only one q-point would be calculated
        fc.calculate_qpoint_phonon_modes(
            qpts, dipole_parameter=dipole_parameter, reduce_qpts=False,
            **calc_modes_kwargs)
        end = time.time()
        t_per_qpt[i] = (end - start)/n
        if print_to_terminal:
            print((sfmt + ': ' + tfmt + ' ms\n').format(
                'Time/qpt', t_per_qpt[i]*1000))
    opt = np.argmin(t_per_qpt)
    if print_to_terminal:
        print('******************************')
        print(('Suggested optimum dipole_parameter is ' + dparamfmt).format(
            dipole_parameters[opt]))
        print((sfmt + ': ' + tfmt + ' s').format('Initialisation Time', t_init[opt]))
        print((sfmt + ': ' + tfmt + ' ms\n').format(
            'Time/qpt', t_per_qpt[opt]*1000))

    return (dipole_parameters[opt], t_init[opt], t_per_qpt[opt],
            dipole_parameters, t_init, t_per_qpt)


def get_parser() -> ArgumentParser:
    parser, sections = _get_cli_parser(
        features={'read-fc', 'dipole-parameter-optimisation'})
    parser.description=(
        'Run and time an interpolation calculation for a small number of '
        'q-points for different values of dipole_parameter to determine '
        'dipole_parameter\'s optimum value for this material')

    return parser



import math
from typing import Dict, Optional, Union, TypeVar, Any, Type
from collections.abc import Mapping

import numpy as np

from euphonic.validate import _check_constructor_inputs
from euphonic.io import _obj_from_json_file, _obj_to_dict, _process_dict
from euphonic.readers import castep, phonopy
from euphonic.util import (direction_changed, is_gamma, get_reference_data)
from euphonic import (ureg, Quantity, Crystal, DebyeWaller, QpointFrequencies,
                      StructureFactor, Spectrum1DCollection)


class QpointPhononModes(QpointFrequencies):
    """
    A class to read and store vibrational data from model (e.g. CASTEP)
    output files

    Attributes
    ----------
    crystal
        Lattice and atom information
    n_qpts
        Number of q-points in the object
    qpts
        Shape (n_qpts, 3) float ndarray. Q-point coordinates, in
        fractional coordinates of the reciprocal lattice
    weights
        Shape (n_qpts,) float ndarray. The weight for each q-point,
        for Brillouin Zone integration over symmetry-reduced q-points
    frequencies
        Shape (n_qpts, 3*crystal.n_atoms) float Quantity. Phonon
        frequencies per q-point and mode
    eigenvectors
        Shape (n_qpts, 3*crystal.n_atoms, crystal.n_atoms, 3) complex
        ndarray. The dynamical matrix eigenvectors in Cartesian
        coordinates, using the same Cartesian basis as the
        cell_vectors in the crystal object
    """
    T = TypeVar('T', bound='QpointPhononModes')

    def __init__(self, crystal: Crystal, qpts: np.ndarray,
                 frequencies: Quantity, eigenvectors: np.ndarray,
                 weights: Optional[np.ndarray] = None) -> None:
        """
        Parameters
        ----------
        crystal
            Lattice and atom information
        qpts
            Shape (n_qpts, 3) float ndarray. Q-point coordinates, in
            fractional coordinates of the reciprocal lattice
        frequencies
            Shape (n_qpts, 3*crystal.n_atoms) float Quantity. Phonon
            frequencies per q-point and mode
        eigenvectors
            Shape (n_qpts, 3*crystal.n_atoms, crystal.n_atoms, 3)
            complax ndarray. The dynamical matrix eigenvectors in
            Cartesian coordinates, using the same Cartesian basis
            as the cell_vectors in the crystal object
        weights
            Shape (n_qpts,) float ndarray. The weight for each q-point,
            for Brillouin Zone integration over symmetry-reduced
            q-points. If None, equal weights are assumed
        """
        super().__init__(crystal, qpts, frequencies, weights)
        n_qpts = len(qpts)
        n_at = crystal.n_atoms
        # Check freqs axis 1 shape here - QpointFrequencies doesn't
        # enforce that the number of modes = 3*(number of atoms)
        _check_constructor_inputs(
            [frequencies, eigenvectors],
            [Quantity, np.ndarray],
            [(n_qpts, 3*n_at), (n_qpts, 3*n_at, n_at, 3)],
            ['frequencies', 'eigenvectors'])
        self.eigenvectors = eigenvectors

    def reorder_frequencies(self,
                            reorder_gamma: bool = True) -> None:
        """
        By doing a dot product of eigenvectors at adjacent q-points,
        determines which modes are most similar and reorders the
        frequencies at each q-point. This means that the same mode
        will have the same index across different q-points, so will
        be plotted as the same colour in a dispersion plot, and can
        be followed across q-space.

        Parameters
        ----------
        reorder_gamma
            Whether to reorder frequencies at gamma-equivalent points.
            If an analytical correction has been applied at the gamma
            points (i.e LO-TO splitting) mode assignments can be
            incorrect at adjacent q-points where the correction hasn't
            been applied. So you might not want to reorder at gamma for
            some materials
        """
        n_qpts = self.n_qpts
        n_branches = 3*self.crystal.n_atoms
        qpts = self.qpts
        eigenvecs = self.eigenvectors

        # Initialise map, don't reorder first q-point
        mode_map = np.zeros((n_qpts, n_branches), dtype=np.int32)
        mode_map[0] = np.arange(n_branches)

        # Only calculate reordering if the direction hasn't
        # changed
        calc_reorder = np.concatenate(([True], np.logical_not(
            direction_changed(qpts))))

        if not reorder_gamma:
            gamma_i = np.where(is_gamma(qpts))[0]
            # Don't reorder at gamma
            calc_reorder[gamma_i[gamma_i > 0] - 1] = False
            # Or at the point after
            calc_reorder[gamma_i[gamma_i < n_qpts - 1]] = False

        for i in range(1, n_qpts):
            # Compare eigenvectors for each mode for this q-point with
            # every mode for the previous q-point
            # Explicitly broadcast arrays with repeat and tile to ensure
            # correct multiplication of modes
            curr_evecs = eigenvecs[i, :, :, :]
            prev_evecs = eigenvecs[i - 1, :, :, :]
            current_eigenvecs = np.repeat(curr_evecs, n_branches, axis=0)
            prev_eigenvecs = np.tile(prev_evecs, (n_branches, 1, 1))

            if calc_reorder[i-1]:
                # Compute complex conjugated dot product of every mode
                # of this q-point with every mode of previous q-point,
                # and sum the dot products over atoms (i.e. multiply
                # eigenvectors elementwise, then sum over the last 2
                # dimensions)
                dots = np.absolute(np.einsum('ijk,ijk->i',
                                             np.conj(prev_eigenvecs),
                                             current_eigenvecs))

                # Create matrix of dot products for each mode of this
                # q-point with each mode of the previous q-point
                dot_mat = np.reshape(dots, (n_branches, n_branches))

                # Find greatest dot product
                for j in range(n_branches):
                    max_i = (np.argmax(dot_mat))
                    # Modes are dot_mat rows
                    mode = int(max_i/n_branches)
                    # Prev q-pt modes are columns
                    prev_mode = max_i % n_branches
                    # Ensure modes aren't mapped more than once
                    dot_mat[mode, :] = 0
                    dot_mat[:, prev_mode] = 0

                    prev_mode_idx = np.where(mode_map[i-1] == prev_mode)[0][0]
                    mode_map[i, prev_mode_idx] = mode
            else:
                mode_map[i] = mode_map[i-1]

        # Actually rearrange frequencies/eigenvectors
        for i in range(n_qpts):
            self._frequencies[i] = self._frequencies[i, mode_map[i]]
            evec_tmp = np.copy(self.eigenvectors[i, mode_map[i]])
            self.eigenvectors[i] = evec_tmp

    def calculate_structure_factor(
        self,
        scattering_lengths: Union[str, Dict[str, Quantity]] = 'Sears1992',
        dw: Optional[DebyeWaller] = None,
        ) -> StructureFactor:
        """
        Calculate the one phonon inelastic scattering for neutrons at
        each q-point

        Parameters
        ----------
        scattering_lengths
            Dataset of coherent scattering length for each
            element in the structure. This may be provided in 3 ways:

            - A string naming an appropriate data collection packaged with
              Euphonic (including the default value 'Sears1992'). This will be
              passed to the ``collection`` argument of
              :obj:`euphonic.util.get_reference_data()`.

            - A string filename for a user's customised data file in the same
              format as those packaged with Euphonic.

            - An explicit dictionary of float Quantity, giving spin- and
              isotope-averaged coherent scattering length for each element in
              the structure, e.g.::

                {'O': 5.803*ureg('fm'), 'Zn': 5.680*ureg('fm')}

        dw
            Data for thermal motion effects. Typically this is computed over a
            converged Monkhort-Pack grid, which need not correspond to the
            q-points of this QpointPhononModes object.

        Returns
        -------
        sf
            An object containing the structure factor for each q-point
            and phonon mode

        Notes
        -----

        This method calculates the mode-resolved (not binned in energy)
        coherent one-phonon neutron scattering function
        :math:`S(Q, \\omega_{\\mathbf{q}\\nu})` **per atom**, as defined in
        [1]_:

        .. math::

          S(Q, \\omega_{\\mathbf{q}\\nu}) =
              \\frac{1}{2N_{atom}} \\
              \\left\\lvert
              {\\sum_{\\kappa}{\\frac{b_\\kappa}{M_{\\kappa}^{1/2}{\\omega_{\\mathbf{q}\\nu}^{1/2}}}
              (\\mathbf{Q}\\cdot \\mathbf{e}_{\\mathbf{q}\\nu\\kappa})\\exp\\left(i\\mathbf{Q}{\\cdot}\\mathbf{r}_{\\kappa}\\right)
              \\exp\\left(-W_{\\kappa}\\right)}}
              \\right\\rvert^2

        Where :math:`\\nu` runs over phonon modes, :math:`\\kappa` runs
        over atoms, :math:`b_\\kappa` is the coherent neutron
        scattering length, :math:`M_{\\kappa}` is the atom mass,
        :math:`r_{\\kappa}` is the vector from the origin to atom
        :math:`\\kappa` in the unit cell,
        :math:`\\mathbf{e}_{\\mathbf{q}\\nu\\kappa}` are the normalised
        Cartesian eigenvectors, :math:`\\omega_{\\mathbf{q}\\nu}` are
        the frequencies and :math:`\\exp\\left(-W_\\kappa\\right)` is the
        Debye-Waller factor. :math:`N_{atom}` is the number of atoms in
        the unit cell, so the returned structure factor is **per atom**
        of sample.

        .. [1] M.T. Dove, Structure and Dynamics, Oxford University Press, Oxford, 2003, 225-226

        """
        if isinstance(scattering_lengths, str):
            scattering_length_data = get_reference_data(
                collection=scattering_lengths,
                physical_property='coherent_scattering_length')
        elif isinstance(scattering_lengths, dict):
            scattering_length_data = scattering_lengths
        else:
            raise TypeError((
                f'Unexpected type for scattering_lengths, should be str '
                f'or dict, got {type(scattering_lengths)}'))

        sl = [scattering_length_data[x].to('bohr').magnitude
              for x in self.crystal.atom_type]

        # Calculate normalisation factor
        norm_factor = sl/np.sqrt(self.crystal._atom_mass)

        # Calculate the exp factor for all atoms and qpts. atom_r is in
        # fractional coords, so Qdotr = 2pi*qh*rx + 2pi*qk*ry...
        exp_factor = np.exp(1J*2*math.pi*np.einsum(
            'ij,kj->ik', self.qpts, self.crystal.atom_r))

        # Eigenvectors are in Cartesian so need to convert hkl to
        # Cartesian by computing dot with hkl and reciprocal lattice
        recip = self.crystal.reciprocal_cell().to('1/bohr').magnitude
        Q = np.einsum('ij,jk->ik', self.qpts, recip)

        # Calculate dot product of Q and eigenvectors for all branches
        # atoms and q-points
        eigenv_dot_q = np.einsum('ijkl,il->ijk', np.conj(self.eigenvectors), Q)

        # Calculate Debye-Waller factors
        temperature = None
        if dw:
            temperature = dw.temperature
            if dw.crystal.n_atoms != self.crystal.n_atoms:
                raise ValueError((
                    'The DebyeWaller object used as dw is not '
                    'compatible with the QPointPhononModes object (they'
                    ' have a different number of atoms)'))
            dw_factor = np.exp(-np.einsum('jkl,ik,il->ij',
                                          dw._debye_waller, Q, Q))
            exp_factor *= dw_factor

        # Multiply Q.eigenvector, exp factor and normalisation factor
        term = np.einsum('ijk,ik,k->ij', eigenv_dot_q, exp_factor, norm_factor)

        # Take mod squared and divide by frequency to get intensity
        sf = np.real(
            np.absolute(term*np.conj(term))/np.absolute(self._frequencies))
        sf /= 2*self.crystal.n_atoms

        return StructureFactor(
            self.crystal, self.qpts, self.frequencies,
            sf*ureg('bohr**2').to('mbarn'),
            temperature=temperature)

    def calculate_debye_waller(
        self, temperature: Quantity,
        frequency_min: Quantity = Quantity(0.01, 'meV'),
        symmetrise: bool = True) -> DebyeWaller:
        """
        Calculate the 3 x 3 Debye-Waller exponent for each atom over the
        q-points contained in this object

        Parameters
        ----------
        temperature
            Scalar float Quantity. The temperature to use in the
            Debye-Waller calculation
        frequency_min
            Scalar float Quantity in energy units. Excludes frequencies below
            this limit from the calculation, as the calculation contains a
            1/frequency factor which would result in infinite values. This
            also allows negative frequencies to be excluded
        symmetrise
            Whether to symmetrise the Debye-Waller factor based on the
            crystal symmetry operations. Note that if the Debye-Waller
            exponent is not symmetrised, the results may not be the
            same for unfolded and symmetry-reduced q-point grids

        Returns
        -------
        dw
            An object containing the 3x3 Debye-Waller exponent matrix
            for each atom

        Notes
        -----

        As part of the structure factor calculation,
        :math:`\\exp\\left(W_\\kappa\\right)` is the anisotropic
        Debye-Waller factor for atom :math:`\\kappa`, and the exponent can
        be written as:

        .. math::

          W_\\kappa =
            \\sum\\limits_{\\alpha\\beta}W^{\\alpha\\beta}_{\\kappa}
              Q_{\\alpha}Q_{\\beta}

        This function calculates the :math:`W^{\\alpha\\beta}_{\\kappa}`
        part of the exponent, which is  a
        :math:`N_\\mathrm{atom}\\times3\\times3` matrix that is
        independent of Q, so for efficiency can be precalculated to be
        used in the structure factor calculation. The Debye-Waller
        exponent matrix is calculated by [2]_:

        .. math::

          W^{\\alpha\\beta}_{\\kappa} =
          \\frac{\\hbar}{4M_{\\kappa}\\sum\\limits_{\\mathbf{q}}{\\mathrm{weight}_\\mathbf{q}}}
          \\sum\\limits_{\\mathbf{q}\\nu \\in{BZ}}\\mathrm{weight}_\\mathbf{q}\\frac{e_{\\mathbf{q}\\nu\\kappa\\alpha}e^{*}_{\\mathbf{q}\\nu\\kappa\\beta}}
          {\\omega_{\\mathbf{q}\\nu}}
          \\mathrm{coth}\\left(\\frac{\\hbar\\omega_{\\mathbf{q}\\nu}}{2k_BT}\\right)

        Where the sum is over q-points and modes :math:`\\nu` in the
        first Brillouin Zone (BZ), :math:`\\kappa` runs over atoms,
        :math:`\\alpha,\\beta` run over the Cartesian directions,
        :math:`M_{\\kappa}` is the atom mass,
        :math:`e_{q\\nu\\kappa\\alpha}` are the scalar components of the
        normalised Cartesian eigenvectors, :math:`\\omega_{\\mathbf{q}\\nu}`
        are the frequencies, and :math:`\\mathrm{weight}_\\mathbf{q}`
        is the per q-point symmetry weight (if the q-points are not
        symmetry-reduced, all weights will be equal).

        .. [2] G.L. Squires, Introduction to the Theory of Thermal Neutron Scattering, Dover Publications, New York, 1996, 34-37
        """

        # Convert units
        kB = (1*ureg.k).to('hartree/K').magnitude
        n_atoms = self.crystal.n_atoms
        atom_mass = self.crystal._atom_mass
        freqs = self._frequencies
        freq_min = frequency_min.to('hartree').magnitude
        qpts = self.qpts
        evecs = self.eigenvectors
        weights = self.weights
        temp = temperature.to('K').magnitude

        mass_term = 1/(4*atom_mass)

        # Mask out frequencies below frequency_min
        freq_mask = np.ones(freqs.shape)
        freq_mask[freqs < freq_min] = 0

        if temp > 0:
            x = freqs/(2*kB*temp)
            freq_term = 1/(freqs*np.tanh(x))
        else:
            freq_term = 1/(freqs)
        dw = np.zeros((n_atoms, 3, 3))
        # Calculating the e.e* term is expensive, do in chunks
        chunk = 1000
        for i in range(int((len(qpts) - 1)/chunk) + 1):
            qi = i*chunk
            qf = min((i + 1)*chunk, len(qpts))

            evec_term = np.real(
                np.einsum('ijkl,ijkm->ijklm',
                          evecs[qi:qf],
                          np.conj(evecs[qi:qf])))

            dw += (np.einsum('i,k,ij,ij,ijklm->klm',
                             weights[qi:qf], mass_term, freq_term[qi:qf],
                             freq_mask[qi:qf], evec_term))

        dw = dw/np.sum(weights)
        if symmetrise:
            dw_tmp = np.zeros(dw.shape)
            (rot, trans,
             eq_atoms) = self.crystal.get_symmetry_equivalent_atoms()
            cell_vec = self.crystal._cell_vectors
            recip_vec = self.crystal.reciprocal_cell().to('1/bohr').magnitude
            rot_cart = np.einsum('ijk,jl,km->ilm',
                                 rot, cell_vec, recip_vec)/(2*np.pi)
            for s in range(len(rot)):
                dw_tmp[eq_atoms[s]] += np.einsum('ij,kjl,ml->kim',
                                                 rot_cart[s], dw, rot_cart[s])
            dw = dw_tmp/len(rot)

        dw = dw*ureg('bohr**2').to(self.crystal.cell_vectors_unit + '**2')
        return DebyeWaller(self.crystal, dw, temperature)

    def calculate_pdos(
            self, dos_bins: Quantity,
            mode_widths: Optional[Quantity] = None,
            mode_widths_min: Quantity = Quantity(0.01, 'meV'),
            adaptive_method: Optional[str] = 'reference',
            adaptive_error: Optional[float] = 0.01,
            weighting: Optional[str] = None,
            cross_sections: Union[str, Dict[str, Quantity]] = 'BlueBook',
            ) -> Spectrum1DCollection:
        """
        Calculates partial density of states for each atom in the unit
        cell

        Parameters
        ----------
        dos_bins
            Shape (n_e_bins + 1,) float Quantity. The energy bin edges
            to use for calculating the DOS.
        mode_widths
            Shape (n_qpts, n_branches) float Quantity in energy units.
            The broadening width for each mode at each q-point, for
            adaptive broadening.
        mode_widths_min
            Scalar float Quantity in energy units. Sets a lower limit on
            the mode widths, as mode widths of zero will result in
            infinitely sharp peaks.
        adaptive_method
            String. Specifies whether to use slow, reference adaptive method or
            faster, approximate method. Allowed options are 'reference'
            or 'fast', default is 'reference'.
        adaptive_error
            Scalar float. Acceptable error for gaussian approximations
            when using the fast adaptive method, defined as the absolute
            difference between the areas of the true and approximate gaussians
        weighting
            One of {'coherent', 'incoherent', 'coherent-plus-incoherent'}.
            If provided, produces a neutron-weighted DOS, weighted by
            either the coherent, incoherent, or sum of coherent and
            incoherent neutron scattering cross-sections.
        cross_sections
            A dataset of cross-sections for each element in the structure,
            it can be a string specifying a dataset, or a dictionary
            explicitly giving the cross-sections for each element.

            If cross_sections is a string, it is passed to the ``collection``
            argument of :obj:`euphonic.util.get_reference_data()`. This
            collection must contain the 'coherent_cross_section' or
            'incoherent_cross_section' physical property, depending on
            the ``weighting`` argument. If ``weighting`` is None, this
            string argument is not used.

            If cross sections is a dictionary, the ``weighting`` argument is
            ignored, and these cross-sections are used directly to calculate
            the neutron-weighted DOS. It must contain a key for each element
            in the structure, and each value must be a Quantity in the
            appropriate units, e.g::

                {'La': 8.0*ureg('barn'), 'Zr': 9.5*ureg('barn')}

        Returns
        -------

        dos
            A collection of spectra, with the energy bins on the x-axis
            and PDOS for each atom in the unit cell on the y-axis,
            metadata describes the index and species of each PDOS. If
            the PDOS is not neutron-weighted, it is in units of modes
            per energy unit per atom, such that the integrated area of
            the total DOS is equal to 3. If the PDOS is weighted by the
            neutron cross-section, it is in units of area per unit
            energy per atom, see Notes for details.

        Notes
        -----

        The PDOS is calculated as:

        .. math::

            PDOS_{\\kappa}(\\omega) = \\frac{1}{N_\\mathrm{atom}\\sum\\limits_{\\mathbf{q}}{\\mathrm{weight}_{\\mathbf{q}}}}
                                      \\sum_{\\mathbf{q}\\nu\\alpha\\in{BZ}}
                                      \\mathrm{weight}_{\\mathbf{q}} {e_{\\mathbf{q}\\nu\\kappa\\alpha}e^{*}_{\\mathbf{q}\\nu\\kappa\\alpha}}
                                      \\delta(\\omega - \\omega_{\\mathbf{q}\\nu})

        Where :math:`N_\\mathrm{atom}` is the number of atoms in the unit cell,
        :math:`\\mathrm{weight}_\\mathbf{q}` is the per q-point weight,
        :math:`e_{\\mathbf{q}\\nu\\kappa\\alpha}` is the normalised
        eigenvector at :math:`\\mathbf{q}` for mode :math:`\\nu`, atom
        :math:`\\kappa` in Cartesian direction :math:`\\alpha` and
        :math:`\\omega_{\\mathbf{q}\\nu}` is the phonon frequency at
        :math:`\\mathbf{q}` for mode :math:`\\nu`, and the sum
        :math:`BZ` is over the 1st Brillouin Zone.

        The neutron-weighted PDOS is calculated as:

        .. math::

            PDOS^\\mathrm{neutron}_{\\kappa}(\\omega) = \\frac{\\sigma_{\\kappa}M_\\mathrm{avg}}{M_\\kappa} PDOS_{\\kappa}(\\omega)

        Where :math:`\\sigma_\\kappa` is the neutron scattering cross
        section, :math:`M_\\mathrm{avg}` is the average mass of atoms in the
        unit cell and :math:`M_\\kappa` is the mass of atom
        :math:`\\kappa`, so the neutron-weighted PDOS is returned in
        units of area per unit energy.

        """
        weighting_opts = [None, 'coherent', 'incoherent',
                          'coherent-plus-incoherent']
        if weighting not in weighting_opts:
            raise ValueError(f'Invalid value for weighting, got '
                             f'{weighting}, should be one of '
                             f'{weighting_opts}')

        cross_sections_data = None
        if isinstance(cross_sections, str):
            if weighting is not None:
                if weighting == 'coherent-plus-incoherent':
                    weights = ['coherent', 'incoherent']
                else:
                    weights = [weighting]
                cross_sections_data = [get_reference_data(
                    collection=cross_sections,
                    physical_property=f'{weight}_cross_section')
                                       for weight in weights]
        elif isinstance(cross_sections, Mapping):
            cross_sections_data = [cross_sections]
        else:
            raise TypeError(f'Unexpected type for cross_sections, expected '
                            f'str or dict, got {type(cross_sections)}')

        if cross_sections_data is not None:
            cs = [cross_sections_data[0][x] for x in self.crystal.atom_type]
            if len(cross_sections_data) == 2:
                cs2 = [cross_sections_data[1][x]
                        for x in self.crystal.atom_type]
                cs = [sum(x) for x in zip(cs, cs2)]
            # Account for cross sections in different, or invalid, units
            ex_units = '[length]**2'
            if not cs[0].check(ex_units):
                raise ValueError((
                    f'Unexpected dimensionality in cross_sections units, '
                    f'expected {ex_units}, got {str(cs[0].dimensionality)}'))
            cs = [x.to('mbarn').magnitude for x in cs]*ureg('mbarn')
        else:
            cs = None

        evec_weights = np.real(np.einsum('ijkl,ijkl->ijk',
                                         self.eigenvectors,
                                         np.conj(self.eigenvectors)))
        # Normalise DOS to per atom
        evec_weights /= self.crystal.n_atoms
        crystal = self.crystal
        for i in range(crystal.n_atoms):
            dos = self._calculate_dos(dos_bins, mode_widths=mode_widths,
                                      mode_widths_min=mode_widths_min,
                                      adaptive_method=adaptive_method,
                                      adaptive_error=adaptive_error,
                                      mode_weights=evec_weights[:, :, i])
            if cs is not None:
                # Neutron weighted DOS is per atom of sample
                avg_atom_mass = np.mean(crystal._atom_mass)
                dos *= cs[i]*avg_atom_mass/crystal._atom_mass[i]
            if i == 0:
                all_dos_y_data = np.zeros((
                    crystal.n_atoms, len(dos_bins) - 1))*dos.units
            all_dos_y_data[i] = dos
        metadata = {'line_data':
            [{'species': symbol, 'index': idx}
             for idx, symbol in enumerate(crystal.atom_type)]}
        if weighting is not None:
            metadata['weighting'] = weighting

        return Spectrum1DCollection(
            dos_bins, all_dos_y_data, metadata=metadata)

    def to_dict(self) -> Dict[str, Any]:
        """
        Convert to a dictionary. See QpointPhononModes.from_dict for
        details on keys/values
        """
        dout = _obj_to_dict(self, ['crystal', 'n_qpts', 'qpts', 'frequencies',
                                   'eigenvectors', 'weights'])
        return dout

    def to_qpoint_frequencies(self) -> QpointFrequencies:
        """
        Create a QpointFrequencies object
        """
        return QpointFrequencies(
            self.crystal, self.qpts, self.frequencies, self.weights)

    @classmethod
    def from_dict(cls: Type[T], d: Dict[str, Any]) -> T:
        """
        Convert a dictionary to a QpointPhononModes object

        Parameters
        ----------
        d : dict
            A dictionary with the following keys/values:

            - 'crystal': dict, see Crystal.from_dict
            - 'qpts': (n_qpts, 3) float ndarray
            - 'frequencies': (n_qpts, 3*crystal.n_atoms) float ndarray
            - 'frequencies_unit': str
            - 'eigenvectors': (n_qpts, 3*crystal.n_atoms, crystal.n_atoms, 3) complex ndarray

            There are also the following optional keys:

            - 'weights': (n_qpts,) float ndarray
        """
        crystal = Crystal.from_dict(d['crystal'])
        d = _process_dict(d, quantities=['frequencies'], optional=['weights'])
        return cls(crystal, d['qpts'], d['frequencies'],
                   d['eigenvectors'], d['weights'])

    @classmethod
    def from_json_file(cls: Type[T], filename: str) -> T:
        """
        Read from a JSON file. See QpointPhononModes.from_dict for
        required fields

        Parameters
        ----------
        filename
            The file to read from
        """
        return _obj_from_json_file(cls, filename,
                                   type_dict={'eigenvectors': np.complex128})

    @classmethod
    def from_castep(cls: Type[T], filename: str,
                    average_repeat_points: bool = True) -> T:
        """
        Reads precalculated phonon mode data from a CASTEP .phonon file

        Parameters
        ----------
        filename
            The path and name of the .phonon file to read
        average_repeat_points
            If multiple frequency/eigenvectors blocks are included with the
            same q-point index (i.e. for Gamma-point with LO-TO splitting),
            scale the weights such that these sum to the given weight
        """
        data = castep.read_phonon_data(
            filename,
            average_repeat_points=average_repeat_points)

        return cls.from_dict(data)

    @classmethod
    def from_phonopy(cls: Type[T], path: str = '.',
                     phonon_name: str = 'band.yaml',
                     phonon_format: Optional[str] = None,
                     summary_name: str = 'phonopy.yaml') -> T:
        """
        Reads precalculated phonon mode data from a Phonopy
        mesh/band/qpoints.yaml/hdf5 file. May also read from
        phonopy.yaml for structure information.

        Parameters
        ----------
        path
            Path to directory containing the file(s)
        phonon_name
            Name of Phonopy file including the frequencies
        phonon_format
            Format of the phonon_name file if it isn't obvious from the
            phonon_name extension, one of {'yaml', 'hdf5'}
        summary_name
            Name of Phonopy summary file to read the crystal information
            from. Crystal information in the phonon_name file takes
            priority, but if it isn't present, crystal information is
            read from summary_name instead
        """
        data = phonopy.read_phonon_data(
            path=path, phonon_name=phonon_name, phonon_format=phonon_format,
            summary_name=summary_name)
        return cls.from_dict(data)



from collections import defaultdict
import re
import struct
from typing import Dict, Any, TextIO, BinaryIO, Tuple, Optional, List, Union

import numpy as np

from euphonic import ureg


def read_phonon_dos_data(
        filename: str,
        cell_vectors_unit: str = 'angstrom',
        atom_mass_unit: str = 'amu',
        frequencies_unit: str = 'meV',
        mode_gradients_unit: str = 'meV*angstrom') -> Dict[str, Any]:
    """
    Reads data from a .phonon_dos file and returns it in a dictionary

    Parameters
    ----------
    filename
        The path and name of the .phonon file to read
    cell_vectors_unit
        The unit to return the cell vectors in
    atom_mass_unit
        The unit to return the atom masses in
    frequencies_unit
        The unit to return the frequencies in

    Returns
    -------
    data_dict : dict
        A dict with the following keys: 'n_atoms', 'cell_vectors',
        'cell_vectors_unit', 'atom_r', 'atom_type', 'atom_mass',
        'atom_mass_unit', 'qpts', 'weights', 'frequencies',
        'frequencies_unit', 'mode_gradients', 'mode_gradients_unit',
        'dos_bins', 'dos_bins_unit', 'dos', 'pdos'
    """
    with open(filename, 'r') as f:

        f.readline()  # Skip BEGIN header
        n_atoms = int(f.readline().split()[-1])
        _ = int(f.readline().split()[-1])  # n_species is not used
        n_branches = int(f.readline().split()[-1])
        n_bins = int(f.readline().split()[-1])
        (cell_vectors, atom_r, atom_type, atom_mass) = _read_crystal_info(
            f, n_atoms)
        f.readline()  # Skip BEGIN GRADIENTS

        qpts = np.empty((0, 3))
        weights = np.empty((0,))
        freqs = np.empty((0, n_branches))
        mode_grads = np.empty((0, n_branches))
        while True:
            try:
                _, qpt, qweight, qfreq, qmode_grad = _read_frequency_block(
                    f, n_branches, extra_columns=[0])
            except IndexError:
                # This should indicate we've reached 'END GRADIENTS' line
                break
            qpts = np.concatenate((qpts, [qpt]))
            weights = np.concatenate((weights, [qweight]))
            freqs = np.concatenate((freqs, [qfreq]))
            mode_grads = np.concatenate((mode_grads, [qmode_grad.squeeze()]))
        line = f.readline()
        if 'BEGIN DOS' not in line:
            raise RuntimeError(
                f'Expected "BEGIN DOS" in {filename}, got {line}')
        # max_rows arg not available until Numpy 1.16.0
        try:
            dos_data = np.loadtxt(f, max_rows=n_bins)
        except TypeError:
            data = f.readlines()
            dos_data = np.array([[float(elem) for elem in line.split()]
                                 for line in data[:n_bins]])

    data_dict: Dict[str, Any] = {}
    data_dict['crystal'] = {}
    cry_dict = data_dict['crystal']
    cry_dict['n_atoms'] = n_atoms
    cry_dict['cell_vectors'] = (cell_vectors*ureg('angstrom').to(
        cell_vectors_unit)).magnitude
    cry_dict['cell_vectors_unit'] = cell_vectors_unit
    cry_dict['atom_r'] = atom_r
    cry_dict['atom_type'] = atom_type
    cry_dict['atom_mass'] = atom_mass*(ureg('amu')).to(
        atom_mass_unit).magnitude
    cry_dict['atom_mass_unit'] = atom_mass_unit

    data_dict['qpts'] = qpts
    data_dict['weights'] = weights
    data_dict['frequencies'] = ((freqs*(1/ureg.cm)).to(
        frequencies_unit)).magnitude
    data_dict['frequencies_unit'] = frequencies_unit
    # Pint conversion 'angstrom*(1/cm)' to 'angstrom*(meV)' doesn't
    # work so convert separately
    mode_grads = mode_grads*ureg('angstrom').to(cell_vectors_unit).magnitude
    data_dict['mode_gradients'] = mode_grads*ureg('1/cm').to(
            frequencies_unit).magnitude
    data_dict['mode_gradients_unit'] = (frequencies_unit + '*'
                                        + cell_vectors_unit)
    data_dict['dos_bins'] = dos_data[:, 0]*ureg('1/cm').to(
            frequencies_unit).magnitude
    data_dict['dos_bins_unit'] = frequencies_unit

    data_dict['dos'] = {}
    data_dict['dos_unit'] = frequencies_unit
    # Avoid issues in converting DOS, Pint allows
    # cm^-1 -> meV but not cm -> 1/meV
    dos_conv = (1*ureg('1/cm').to(frequencies_unit)).magnitude
    dos_dict = data_dict['dos']
    dos_dict['Total'] = dos_data[:, 1]/dos_conv
    _, idx = np.unique(atom_type, return_index=True)
    unique_types = atom_type[np.sort(idx)]
    for i, species in enumerate(unique_types):
        dos_dict[species] = dos_data[:, i + 2]/dos_conv

    return data_dict


def read_phonon_data(
        filename: str,
        cell_vectors_unit: str = 'angstrom',
        atom_mass_unit: str = 'amu',
        frequencies_unit: str = 'meV',
        read_eigenvectors: bool = True,
        average_repeat_points: bool = True) -> Dict[str, Any]:
    """
    Reads data from a .phonon file and returns it in a dictionary

    Parameters
    ----------
    filename
        The path and name of the .phonon file to read
    cell_vectors_unit
        The unit to return the cell vectors in
    atom_mass_unit
        The unit to return the atom masses in
    frequencies_unit
        The unit to return the frequencies in
    read_eigenvectors
        Whether to read the eigenvectors and return them in the
        dictionary
    average_repeat_points
        If multiple frequency/eigenvectors blocks are provided with the same
        q-point index (i.e. for Gamma-point with LO-TO splitting), scale the
        weights such that these sum to the given weight

    Returns
    -------
    data_dict : dict
        A dict with the following keys: 'n_atoms', 'cell_vectors',
        'cell_vectors_unit', 'atom_r', 'atom_type', 'atom_mass',
        'atom_mass_unit', 'qpts', 'weights', 'frequencies',
        'frequencies_unit'

        If read_eigenvectors is True, there will also be an
        'eigenvectors' key
    """
    with open(filename, 'r') as f:

        f.readline()  # Skip BEGIN header
        n_atoms = int(f.readline().split()[-1])
        n_branches = int(f.readline().split()[-1])
        n_qpts = int(f.readline().split()[-1])
        (cell_vectors, atom_r, atom_type, atom_mass) = _read_crystal_info(
            f, n_atoms)

        qpts = np.zeros((n_qpts, 3))
        weights = np.zeros(n_qpts)
        freqs = np.zeros((n_qpts, n_branches))
        if read_eigenvectors:
            eigenvecs = np.zeros((n_qpts, n_branches, n_atoms, 3),
                                 dtype='complex128')

        # Need to loop through file using while rather than number of
        # q-points as points are duplicated on LO-TO splitting
        idx = 0
        previous_qpt_id = -1
        repeated_qpt_ids = defaultdict(set)
        while True:
            try:
                qpt_id, qpt, qweight, qfreq, _ = _read_frequency_block(
                    f, n_branches)
            except EOFError:
                break

            if average_repeat_points and qpt_id == previous_qpt_id:
                repeated_qpt_ids[qpt_id].update({idx - 1, idx})
            previous_qpt_id = qpt_id

            [f.readline() for x in range(2)]  # Skip 2 label lines
            evec_lines = np.array(
                [f.readline() for x in range(n_atoms*n_branches)])
            if read_eigenvectors:
                evec_lines = np.array([x.split()[2:] for x in evec_lines],
                                      dtype=np.float64)
                lines_i = np.column_stack((
                    [evec_lines[:, 0] + evec_lines[:, 1]*1j,
                     evec_lines[:, 2] + evec_lines[:, 3]*1j,
                     evec_lines[:, 4] + evec_lines[:, 5]*1j]))
                qeigenvec = np.zeros((n_branches, n_atoms, 3),
                                     dtype=np.complex128)
                for i in range(n_branches):
                    qeigenvec[i, :, :] = lines_i[i*n_atoms:(i+1)*n_atoms, :]
            # Sometimes there are more than n_qpts q-points in the file
            # due to LO-TO splitting
            if idx < len(qpts):
                qpts[idx] = qpt
                freqs[idx] = qfreq
                weights[idx] = qweight
                if read_eigenvectors:
                    eigenvecs[idx] = qeigenvec
            else:
                qpts = np.concatenate((qpts, [qpt]))
                freqs = np.concatenate((freqs, [qfreq]))
                weights = np.concatenate((weights, [qweight]))
                if read_eigenvectors:
                    eigenvecs = np.concatenate((eigenvecs, [qeigenvec]))
            idx += 1

    # Multiple qpts with same CASTEP q-pt index: correct weights
    if average_repeat_points:
        for qpt_id, indices in repeated_qpt_ids.items():
            indices = np.asarray(list(indices), dtype=int)
            weights[indices] /= indices.size

    data_dict: Dict[str, Any] = {}
    data_dict['crystal'] = {}
    cry_dict = data_dict['crystal']
    cry_dict['n_atoms'] = n_atoms
    cry_dict['cell_vectors'] = (cell_vectors*ureg('angstrom').to(
        cell_vectors_unit)).magnitude
    cry_dict['cell_vectors_unit'] = cell_vectors_unit
    cry_dict['atom_r'] = atom_r
    cry_dict['atom_type'] = atom_type
    cry_dict['atom_mass'] = atom_mass*(ureg('amu')).to(
        atom_mass_unit).magnitude
    cry_dict['atom_mass_unit'] = atom_mass_unit
    data_dict['qpts'] = qpts
    data_dict['weights'] = weights
    data_dict['frequencies'] = ((freqs*(1/ureg.cm)).to(
        frequencies_unit)).magnitude
    data_dict['frequencies_unit'] = frequencies_unit
    if read_eigenvectors:
        data_dict['eigenvectors'] = eigenvecs

    return data_dict


def _read_crystal_info(
        f: TextIO, n_atoms: int
        ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Reads the header crystal information from a CASTEP text file, from
    'Unit cell vectors' to 'END header'

    Parameters
    ----------
    f
        File object in read mode for the .phonon file containing the
        data
    n_atoms
        Number of atoms in the cell

    Returns
    -------
    cell_vectors
        Shape (3, 3) float ndarray. The unit cell vectors in Angstroms
    atom_r
        Shape (n_atoms, 3) float ndarray. The fractional position of
        each atom within the unit cell
    atom_type
        Shape (n_atoms,) string ndarray. The chemical symbols of each
        atom in the unit cell
    atom_mass
        Shape (n_atoms,) float ndarray. The mass of each atom in the
        unit cell in atomic mass units
    """
    while 'Unit cell vectors' not in f.readline():
        pass
    cell_vectors = np.array([[float(x) for x in f.readline().split()[0:3]]
                             for i in range(3)])
    f.readline()  # Skip fractional co-ordinates label
    atom_info = np.array([f.readline().split() for i in range(n_atoms)])
    atom_r = np.array([[float(x) for x in y[1:4]] for y in atom_info])
    atom_type = np.array([x[4] for x in atom_info])
    atom_mass = np.array([float(x[5]) for x in atom_info])
    f.readline()  # Skip END header line

    return cell_vectors, atom_r, atom_type, atom_mass


_qpt_index_pattern = re.compile(r'q-pt=\s*(\d+)')
_qpt_float_pattern = re.compile(r'-?\d+\.\d+')


def _read_frequency_block(
        f: TextIO, n_branches: int, extra_columns: Optional[List] = None
        ) -> Tuple[int, np.ndarray, float, np.ndarray, Optional[np.ndarray]]:
    """
    For a single q-point reads the q-point, weight, frequencies
    and optionally any extra columns

    Parameters
    ----------
    f
        File object in read mode for the file containing the data
    extra_columns
        The index(es) of extra columns to read after the frequencies
        column. e.g. to read the first column after frequencies
        extra_columns = [0]. To read the first and second columns after
        frequencies extra_columns = [0, 1]. For reference, in .phonon
        IR intensities = 0, Raman intensities = 1, and in .phonon_dos
        mode gradients = 0

    Returns
    -------
    iqpt
        CASTEP-assigned index of the q-point
    qpt
        Shape (3,) float ndarray. The q-point in reciprocal fractional
        coordinates
    weight
        The weight of this q-point
    frequencies
        Shape (n_modes,) float ndarray. The phonon frequencies for this
        q-point in 1/cm
    extra
        None if extra_columns is None, otherwise a shape
        (len(extra_columns), n_modes) float ndarray
    """
    qpt_line = f.readline()
    if qpt_line == '':
        raise EOFError

    i_qpt = int(re.findall(_qpt_index_pattern, qpt_line)[0])
    floats = re.findall(_qpt_float_pattern, qpt_line)
    qpt = np.array([float(x) for x in floats[:3]])
    qweight = float(floats[3])
    freq_lines = [f.readline().split()
                  for i in range(n_branches)]
    freq_col = 1
    qfreq = np.array([float(line[freq_col]) for line in freq_lines])
    extra: Optional[np.ndarray] = None
    if extra_columns is not None:
        if len(qpt_line.split()) > 6:
            # Indicates a split gamma point in .phonon, so there is an
            # additional column
            extra_columns = [x + 1 for x in extra_columns]
        extra = np.zeros((len(extra_columns), n_branches))
        for i, col in enumerate(extra_columns):
            extra[i] = np.array(
                [float(line[freq_col + col + 1]) for line in freq_lines])
    return i_qpt, qpt, qweight, qfreq, extra


def read_interpolation_data(
        filename: str,
        cell_vectors_unit: str = 'angstrom',
        atom_mass_unit: str = 'amu',
        force_constants_unit: str = 'hartree/bohr**2',
        born_unit: str = 'e',
        dielectric_unit: str = '(e**2)/(bohr*hartree)') -> Dict[str, Any]:
    """
    Reads data from a .castep_bin or .check file and returns it in a
    dictionary

    Parameters
    ----------
    filename
        The path and name of the file to read
    cell_vectors_unit
        The unit to return the cell vectors in
    atom_mass_unit
        The unit to return the atom masses in
    force_constants_unit
        The unit to return the force constants in
    born_unit
        The unit to return the Born charges in
    dielectric_unit
        The unit to return the dielectric permittivity tensor in

    Returns
    -------
    data_dict : dict
        A dict with the following keys: 'n_atoms', 'cell_vectors',
        'cell_vectors_unit', 'atom_r', 'atom_type', 'atom_mass',
        'atom_mass_unit', 'force_constants', 'force_constants_unit',
        'sc_matrix' and 'cell_origins'. Also contains 'born',
        'born_unit', 'dielectric' and 'dielectric_unit' if they are
        present in the .castep_bin or .check file.
    """
    with open(filename, 'rb') as f:
        int_type = '>i4'
        float_type = '>f8'
        header = ''
        first_cell_read = True
        while header.strip() != b'END':
            header = _read_entry(f)
            if header.strip() == b'BEGIN_UNIT_CELL':
                # CASTEP writes the cell twice: the first is the
                # geometry optimised cell, the second is the original
                # cell. We only want the geometry optimised cell.
                if first_cell_read:
                    (n_atoms, cell_vectors, atom_r, atom_mass,
                     atom_type) = _read_cell(f, int_type, float_type)
                    first_cell_read = False
            elif header.strip() == b'FORCE_CON':
                sc_matrix = np.transpose(np.reshape(
                    _read_entry(f, int_type), (3, 3)))
                n_cells_in_sc = int(np.rint(np.absolute(
                    np.linalg.det(sc_matrix))))
                # Transpose and reshape fc so it is indexed [nc, i, j]
                force_constants = np.ascontiguousarray(np.transpose(
                    np.reshape(_read_entry(f, float_type),
                               (n_cells_in_sc, 3*n_atoms, 3*n_atoms)),
                    axes=[0, 2, 1]))
                cell_origins = np.reshape(
                    _read_entry(f, int_type), (n_cells_in_sc, 3))
                _ = _read_entry(f, int_type)  # FC row not used
            elif header.strip() == b'BORN_CHGS':
                born = np.reshape(
                    _read_entry(f, float_type), (n_atoms, 3, 3))
            elif header.strip() == b'DIELECTRIC':
                dielectric = np.transpose(np.reshape(
                    _read_entry(f, float_type), (3, 3)))

    data_dict: Dict[str, Any] = {}
    data_dict['crystal'] = {}
    cry_dict = data_dict['crystal']
    cry_dict['n_atoms'] = n_atoms
    cry_dict['cell_vectors'] = cell_vectors*ureg(
        'bohr').to(cell_vectors_unit).magnitude
    cry_dict['cell_vectors_unit'] = cell_vectors_unit
    # Normalise atom coordinates
    cry_dict['atom_r'] = atom_r - np.floor(atom_r)
    cry_dict['atom_type'] = atom_type
    cry_dict['atom_mass'] = atom_mass*ureg(
        'electron_mass').to(atom_mass_unit).magnitude
    cry_dict['atom_mass_unit'] = atom_mass_unit

    # Set entries relating to 'FORCE_CON' block
    try:
        data_dict['force_constants'] = force_constants*ureg(
            'hartree/bohr**2').to(force_constants_unit).magnitude
        data_dict['force_constants_unit'] = force_constants_unit
        data_dict['sc_matrix'] = sc_matrix
        data_dict['cell_origins'] = cell_origins
    except NameError:
        raise RuntimeError((
            f'Force constants matrix could not be found in {filename}. '
            f'\nEnsure PHONON_WRITE_FORCE_CONSTANTS: true and a '
            f'PHONON_FINE_METHOD has been chosen when running CASTEP'))

    # Set entries relating to dipoles
    try:
        data_dict['born'] = born*ureg('e').to(born_unit).magnitude
        data_dict['born_unit'] = born_unit
        data_dict['dielectric'] = dielectric*ureg(
            'e**2/(hartree*bohr)').to(dielectric_unit).magnitude
        data_dict['dielectric_unit'] = dielectric_unit
    except UnboundLocalError:
        pass

    return data_dict


def _read_cell(file_obj: BinaryIO, int_type: str, float_type: str
               ) -> Tuple[int, np.ndarray, np.ndarray,
                          np.ndarray, np.ndarray]:
    """
    Read cell data from a .castep_bin or .check file

    Parameters
    ----------
    file_obj
        File object in read mode for the .castep_bin or .check file
    int_type
        Python struct format string describing the size and endian-ness
        of ints in the file
    float_type
        Python struct format string describing the size and endian-ness
        of floats in the file

    Returns
    -------
    n_atoms
        Number of atoms in the unit cell
    cell_vectors
        Shape (3, 3) float ndarray. The unit cell vectors in bohr
    atom_r
        Shape (n_atoms, 3) float ndarray. The fractional position of
        each atom within the unit cell
    atom_mass
        Shape (n_atoms,) float ndarray. The mass of each atom in the
        unit cell in units of electron mass
    atom_type
        Shape (n_atoms,) string ndarray. The chemical symbols of each
        atom in the unit cell
    """
    header = ''
    while header.strip() != b'END_UNIT_CELL':
        header = _read_entry(file_obj)
        if header.strip() == b'CELL%NUM_IONS':
            n_atoms = _read_entry(file_obj, int_type)
        elif header.strip() == b'CELL%REAL_LATTICE':
            cell_vectors = np.transpose(np.reshape(
                _read_entry(file_obj, float_type), (3, 3)))
        elif header.strip() == b'CELL%NUM_SPECIES':
            n_species = _read_entry(file_obj, int_type)
        elif header.strip() == b'CELL%NUM_IONS_IN_SPECIES':
            n_atoms_in_species = _read_entry(file_obj, int_type)
            if n_species == 1:
                n_atoms_in_species = np.array([n_atoms_in_species])
        elif header.strip() == b'CELL%IONIC_POSITIONS':
            max_atoms_in_species = max(n_atoms_in_species)
            atom_r_tmp = np.reshape(_read_entry(file_obj, float_type),
                                    (n_species, max_atoms_in_species, 3))
        elif header.strip() == b'CELL%SPECIES_MASS':
            atom_mass_tmp = _read_entry(file_obj, float_type)
            if n_species == 1:
                atom_mass_tmp = np.array([atom_mass_tmp])
        elif header.strip() == b'CELL%SPECIES_SYMBOL':
            # Need to decode binary string for Python 3 compatibility
            if n_species == 1:
                atom_type_tmp = [_read_entry(file_obj, 'S8')
                                 .strip().decode('utf-8')]
            else:
                atom_type_tmp = [x.strip().decode('utf-8')
                                 for x in _read_entry(file_obj, 'S8')]
    # Get atom_r in correct form
    # CASTEP stores atom positions as 3D array (3,
    # max_atoms_in_species, n_species) so need to slice data to get
    # correct information
    atom_begin = np.insert(np.cumsum(n_atoms_in_species[:-1]), 0, 0)
    atom_end = np.cumsum(n_atoms_in_species)
    atom_r = np.zeros((n_atoms, 3))
    for i in range(n_species):
        atom_r[atom_begin[i]:atom_end[i], :] = atom_r_tmp[
            i, :n_atoms_in_species[i], :]
    # Get atom_type in correct form
    atom_type = np.array([])
    atom_mass = np.array([])
    for at in range(n_species):
        atom_type = np.append(
            atom_type,
            [atom_type_tmp[at] for i in range(n_atoms_in_species[at])])
        atom_mass = np.append(
            atom_mass,
            [atom_mass_tmp[at] for i in range(n_atoms_in_species[at])])

    return n_atoms, cell_vectors, atom_r, atom_mass, atom_type


def _read_entry(file_obj: BinaryIO, dtype: str = ''
                ) -> Union[str, int, float, np.ndarray]:
    """
    Read a record from a Fortran binary file, including the beginning
    and end record markers and return the data inbetween

    Parameters
    ----------
    f
        File object in read mode for the Fortran binary file
    dtype
        String determining what order and type to unpack the bytes as.
        See 'Format Strings' in Python struct documentation

    Returns
    -------
    data
        Data type returned depends on dtype specified. If dtype is not
        specified, return type is a string. If there is more than one
        element in the record, it is returned as an ndarray of floats or
        integers

    """
    def record_mark_read(file_obj):
        # Read 4 byte Fortran record marker
        rawdata = file_obj.read(4)
        if rawdata == b'':
            raise EOFError(
                'Problem reading binary file: unexpected EOF reached')
        return struct.unpack('>i', rawdata)[0]

    begin = record_mark_read(file_obj)
    if dtype:
        n_bytes = int(dtype[-1])
        n_elems = int(begin/n_bytes)
        if n_elems > 1:
            data = np.fromfile(file_obj, dtype=dtype, count=n_elems)
            if 'i' in dtype:
                data = data.astype(np.int32)
            elif 'f' in dtype:
                data = data.astype(np.float64)
        else:
            if 'i' in dtype:
                data = struct.unpack('>i', file_obj.read(begin))[0]
            elif 'f' in dtype:
                data = struct.unpack('>d', file_obj.read(begin))[0]
            else:
                data = file_obj.read(begin)
    else:
        data = file_obj.read(begin)
    end = record_mark_read(file_obj)
    if begin != end:
        raise IOError("""Problem reading binary file: beginning and end
                         record markers do not match""")

    return data



import inspect
from math import ceil
from typing import List, Tuple, Type, TypeVar, Dict, Any
from collections import OrderedDict

import numpy as np
import spglib

from euphonic.util import _cell_vectors_to_volume, _get_unique_elems_and_idx
from euphonic.validate import _check_constructor_inputs, _check_unit_conversion
from euphonic.io import (_obj_to_json_file, _obj_from_json_file,
                         _obj_to_dict, _process_dict)
from euphonic import ureg, Quantity

class Crystal:
    """
    Stores lattice and atom information

    Attributes
    ----------
    cell_vectors
        Shape (3, 3) float Quantity in length units. Cartesian unit
        cell vectors. cell_vectors[0] = a, cell_vectors[:, 0] = x etc.
    n_atoms
        Number of atoms in the unit cell
    atom_r
        Shape (n_atoms, 3) float ndarray. The fractional position of
        each atom within the unit cell
    atom_type
        Shape (n_atoms,) string ndarray. The chemical symbols of each
        atom in the unit cell
    atom_mass
        Shape (n_atoms,) float Quantity in mass units. The mass of each
        atom in the unit cell
    """
    T = TypeVar('T', bound='Crystal')

    def __init__(self, cell_vectors: Quantity, atom_r: np.ndarray,
                 atom_type: np.ndarray, atom_mass: Quantity) -> None:
        """
        Parameters
        ----------
        cell_vectors
            Shape (3, 3) float Quantity in length units. Cartesian unit
            cell vectors. cell_vectors[0] = a,
            cell_vectors[:, 0] = x etc.
        atom_r
            Shape (n_atoms, 3) float ndarray. The fractional position
            of each atom within the unit cell
        atom_type
            Shape (n_atoms,) string ndarray. The chemical symbols of
            each atom in the unit cell
        atom_mass
            Shape (n_atoms,) float Quantity in mass units. The mass
            of each atom in the unit cell
        """
        n_atoms = len(atom_r)
        # Allow empty structure information, but convert to correct
        # shape/type. This is required here to allow reading from .json
        # files where attrs will have been converted to a list so
        # shape/type information will have been lost
        if n_atoms == 0 and atom_r.shape == (0,):
            atom_r = np.zeros((0, 3), dtype=np.float64)
        if n_atoms == 0 and atom_type.shape == (0,):
            atom_type = np.array([], dtype='<U32')

        _check_constructor_inputs(
            [cell_vectors, atom_r, atom_type, atom_mass],
            [Quantity, np.ndarray, np.ndarray, Quantity],
            [(3, 3), (n_atoms, 3), (n_atoms,), (n_atoms,)],
            inspect.getfullargspec(self.__init__)[0][1:])
        self._cell_vectors = cell_vectors.to(ureg.bohr).magnitude
        self.n_atoms = n_atoms
        self.atom_r = atom_r
        self.atom_type = atom_type
        self._atom_mass = atom_mass.to(ureg.m_e).magnitude

        self.cell_vectors_unit = str(cell_vectors.units)
        self.atom_mass_unit = str(atom_mass.units)
                
    @property
    def cell_vectors(self) -> Quantity:
        return self._cell_vectors*ureg('bohr').to(
            self.cell_vectors_unit)

    @cell_vectors.setter
    def cell_vectors(self, value: Quantity) -> None:
        self.cell_vectors_unit = str(value.units)
        self._cell_vectors = value.to('bohr').magnitude

    @property
    def atom_mass(self) -> Quantity:
        return self._atom_mass*ureg('m_e').to(
            self.atom_mass_unit)

    @atom_mass.setter
    def atom_mass(self, value: Quantity) -> None:
        self.atom_mass_unit = str(value.units)
        self._atom_mass = value.to('m_e').magnitude

    def __setattr__(self, name: str, value: Any) -> None:
        _check_unit_conversion(self, name, value,
                               ['cell_vectors_unit', 'atom_mass_unit'])
        super(Crystal, self).__setattr__(name, value)

    def reciprocal_cell(self) -> Quantity:
        """
        Calculates the reciprocal lattice vectors

        Returns
        -------
        recip
            Shape (3, 3) float Quantity in 1/length units, the
            reciprocal lattice vectors
        """
        cv = self._cell_vectors

        bxc = np.cross(cv[1], cv[2])
        cxa = np.cross(cv[2], cv[0])
        axb = np.cross(cv[0], cv[1])
        vol = self._cell_volume()
        norm = 2*np.pi/vol

        recip = np.array([norm*bxc,
                          norm*cxa,
                          norm*axb])*(1/ureg.bohr)

        return recip.to(1/ureg(self.cell_vectors_unit))

    def cell_volume(self) -> Quantity:
        """
        Calculates the cell volume

        Returns
        -------
        volume
            Scalar float Quantity in length**3 units. The cell volume
        """
        vol = self._cell_volume()*ureg.bohr**3
        return vol.to(ureg(self.cell_vectors_unit)**3)

    def _cell_volume(self) -> float:
        return _cell_vectors_to_volume(self._cell_vectors)

    def get_mp_grid_spec(self,
                         spacing: Quantity = 0.1 * ureg('1/angstrom')
                         ) -> Tuple[int, int, int]:
        """
        Get suggested divisions for Monkhorst-Pack grid

        Determine a mesh for even Monkhorst-Pack sampling of the reciprocal
        cell

        Parameters
        ----------
        spacing
            Scalar float quantity in 1/length units. Maximum
            reciprocal-space distance between q-point samples

        Returns
        -------
        grid_spec
            The number of divisions for each reciprocal lattice vector
        """

        recip_length_unit = spacing.units
        lattice = self.reciprocal_cell().to(recip_length_unit)
        grid_spec = np.linalg.norm(lattice.magnitude, axis=1
                                   ) / spacing.magnitude
        # math.ceil is better than np.ceil because it returns ints
        return tuple([ceil(x) for x in grid_spec])

    def to_spglib_cell(self) -> Tuple[List[List[float]],
                                      List[List[float]],
                                      List[int]]:
        """
        Convert to a 'cell' as defined by spglib

        Returns
        -------
        cell
            cell = (lattice, positions, numbers), where lattice is the
            lattice vectors, positions are the fractional atomic
            positions, and numbers are integers distinguishing the
            atomic species
        """
        _, unique_atoms = np.unique(self.atom_type, return_inverse=True)
        cell = (self.cell_vectors.magnitude.tolist(),
                self.atom_r.tolist(),
                unique_atoms.tolist())
        return cell

    def get_species_idx(self) -> 'OrderedDict[str, np.ndarray]':
        """
        Returns a dictionary of each species and their indices

        Returns
        -------
        species_idx
            An ordered dictionary containing each unique species
            symbol as the keys, and their indices as the values,
            in the same order as they appear in atom_type
        """
        species_dict = _get_unique_elems_and_idx(
            [tuple([at]) for at in self.atom_type])
        # Convert tuples back to string
        return OrderedDict([(str(key[0]), value)
                            for key, value in species_dict.items()])

    def get_symmetry_equivalent_atoms(
            self, tol: Quantity = Quantity(1e-5, 'angstrom')
            ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Returns the rotational and translational symmetry operations
        as obtained by spglib.get_symmetry, and also the equivalent
        atoms that each atom gets mapped onto for each symmetry
        operation

        Parameters
        ----------
        tol
           Scalar float Quantity in length units. The distance
           tolerance, if the distance between atoms is less than
           this, they are considered to be equivalent. This is
           also passed to spglib.get_symmetry as symprec

        Returns
        -------
        rotations
           Shape (n_symmetry_ops, 3, 3) integer np.ndarray. The
           rotational symmetry matrices as returned by
           spglib.get_symmetry
        translations
           Shape (n_symmetry_ops, 3) float np.ndarray. The
           rotational symmetry matrices as returned by
           spglib.get_symmetry
        equivalent_atoms
           Shape (n_symmetry_ops, n_atoms) integer np.ndarray.
           The equivalent atoms for each symmetry operation. e.g.
           equivalent_atoms[s, i] = j means symmetry operation s
           maps atom i to atom j

        """
        tol_calc = tol.to('bohr').magnitude
        symprec = tol.to(self.cell_vectors_unit).magnitude
        # Sometimes if symprec is very low, even the identity
        # symmetry op won't be found, and None will be returned
        # For some reason this can't always be reproduced
        symm = spglib.get_symmetry(self.to_spglib_cell(), symprec=symprec)
        if symm is None:
            raise RuntimeError(f'spglib.get_symmetry returned None with '
                               f'symprec={symprec}. Try increasing tol')
        n_ops = len(symm['rotations'])
        equiv_atoms = np.full((n_ops, self.n_atoms), -1, dtype=np.int32)
        atom_r_symm = (np.einsum('ijk,lk->ilj', symm['rotations'], self.atom_r)
                       + symm['translations'][:, np.newaxis, :])
        atom_r_symm -= np.floor(atom_r_symm + 0.5)

        species_idx = self.get_species_idx()
        for spec, idx in species_idx.items():
            for i in idx:
                atom_r_symm_i = atom_r_symm[:, i, :]
                # Difference between symmetry-transformed atom i and all
                # other atoms of that species for each symmetry operation
                diff_frac = (atom_r_symm_i[:, np.newaxis, :]
                             - self.atom_r[np.newaxis, idx, :])
                diff_frac -= np.floor(diff_frac + 0.5)
                diff_cart = np.einsum('ijk,kl->ijl', diff_frac, self._cell_vectors)
                diff_r = np.linalg.norm(diff_cart, axis=2)
                equiv_idx = np.where(diff_r < tol_calc)
                # There should be one matching atom per symm op
                if not np.array_equal(equiv_idx[0], np.arange(n_ops)):
                    for op_idx, diff_r_op in enumerate(diff_r):
                        equiv_idx_op = np.where(diff_r_op < tol_calc)[0]
                        err_info = (f'for {spec} atom at {self.atom_r[i]} for '
                                    f'symmetry op {op_idx}. Rotation '
                                    f'{symm["rotations"][op_idx]} translation '
                                    f'{symm["translations"][op_idx]}')
                        if len(equiv_idx_op) == 0:
                            raise RuntimeError(f'No equivalent atom found {err_info}')
                        elif len(equiv_idx_op) > 1:
                            raise RuntimeError(f'Multiple equivalent atoms found {err_info}')
                equiv_atoms[:, i] = idx[equiv_idx[1]]

        return symm['rotations'], symm['translations'], equiv_atoms

    def to_dict(self) -> Dict[str, Any]:
        """
        Convert to a dictionary. See Crystal.from_dict for details on
        keys/values

        Returns
        -------
        dict
        """
        dout = _obj_to_dict(self, ['cell_vectors', 'n_atoms', 'atom_r',
                                   'atom_type', 'atom_mass'])
        return dout

    def to_json_file(self, filename: str) -> None:
        """
        Write to a JSON file. JSON fields are equivalent to
        Crystal.from_dict keys

        Parameters
        ----------
        filename
            Name of the JSON file to write to
        """
        _obj_to_json_file(self, filename)

    @classmethod
    def from_dict(cls: Type[T], d: Dict[str, Any]) -> T:
        """
        Convert a dictionary to a Crystal object

        Parameters
        ----------
        d
            A dictionary with the following keys/values:

            - 'cell_vectors': (3, 3) float ndarray
            - 'cell_vectors_unit': str
            - 'atom_r': (n_atoms, 3) float ndarray
            - 'atom_type': (n_atoms,) str ndarray
            - 'atom_mass': (n_atoms,) float np.ndaaray
            - 'atom_mass_unit': str

        Returns
        -------
        crystal
        """
        d = _process_dict(d, quantities=['cell_vectors', 'atom_mass'])
        return cls(d['cell_vectors'], d['atom_r'], d['atom_type'],
                   d['atom_mass'])

    @classmethod
    def from_json_file(cls: Type[T], filename: str) -> T:
        """
        Read from a JSON file. See Crystal.from_dict for required fields

        Parameters
        ----------
        filename
            The file to read from

        Returns
        -------
        crystal
        """
        return _obj_from_json_file(cls, filename)

    @classmethod
    def from_cell_vectors(cls: Type[T], cell_vectors: Quantity) -> T:
        """
        Create a Crystal object from just cell vectors, containing no
        detailed structure information (atomic positions, species,
        masses)

        Parameters
        ----------
        cell_vectors
            Shape (3, 3) float Quantity in length units. Cartesian
            unit cell vectors. cell_vectors[0] = a,
            cell_vectors[:, 0] = x etc.

        Returns
        -------
        crystal
        """
        return cls(cell_vectors,
                   np.array([]), np.array([]),
                   Quantity(np.array([], dtype=np.float64), 'amu'))



from argparse import (ArgumentParser, _ArgumentGroup, Namespace,
                      ArgumentDefaultsHelpFormatter, Action)
import json
import os
import pathlib
import warnings
from typing import (Any, Collection, Dict, List,
                    Sequence, Tuple, Union, Optional)

import numpy as np
from pint import UndefinedUnitError
import seekpath

from euphonic import (Crystal, DebyeWaller, ForceConstants, QpointFrequencies,
                      QpointPhononModes, Spectrum1D, Spectrum1DCollection,
                      Quantity, ureg)
import euphonic.util
Unit = ureg.Unit


def _load_euphonic_json(filename: Union[str, os.PathLike],
                        frequencies_only: bool = False
                        ) -> Union[QpointPhononModes, QpointFrequencies,
                                   ForceConstants]:
    with open(filename, 'r') as f:
        data = json.load(f)

    if 'force_constants' in data:
        return ForceConstants.from_json_file(filename)
    elif 'frequencies' in data:
        if 'eigenvectors' in data and not frequencies_only:
            return QpointPhononModes.from_json_file(filename)
        else:
            return QpointFrequencies.from_json_file(filename)
    else:
        raise ValueError("Could not identify Euphonic data in JSON file.")


def _load_phonopy_file(filename: Union[str, os.PathLike],
                       frequencies_only: bool = False
                       ) -> Union[QpointPhononModes, QpointFrequencies,
                                  ForceConstants]:
    path = pathlib.Path(filename)
    loaded_data = None
    if not frequencies_only:
        try:
            loaded_data = QpointPhononModes.from_phonopy(
                path=path.parent, phonon_name=path.name)
        except (KeyError, RuntimeError):
            # KeyError will be raised if it is actually a force
            # constants file, RuntimeError will be raised if
            # it only contains q-point frequencies (no eigenvectors)
            pass

    # Try to read QpointFrequencies if loading QpointPhononModes has
    # failed, or has been specifically requested with frequencies_only
    if frequencies_only or loaded_data is None:
        try:
            loaded_data = QpointFrequencies.from_phonopy(
                path=path.parent, phonon_name=path.name)
        except KeyError:
            pass

    if loaded_data is None:
        phonopy_kwargs: Dict[str, Union[str, os.PathLike]] = {}
        phonopy_kwargs['path'] = path.parent
        if (path.parent / 'BORN').is_file():
            phonopy_kwargs['born_name'] = 'BORN'
        # Set summary_name and fc_name depending on input file
        if path.suffix == '.hdf5':
            if (path.parent / 'phonopy.yaml').is_file():
                phonopy_kwargs['summary_name'] = 'phonopy.yaml'
                phonopy_kwargs['fc_name'] = path.name
            else:
                raise ValueError("Phonopy force_constants.hdf5 file "
                                 "must be accompanied by phonopy.yaml")
        elif path.suffix == '.yaml':
            phonopy_kwargs['summary_name'] = path.name
            # Assume this is a (renamed?) phonopy.yaml file
            if (path.parent / 'force_constants.hdf5').is_file():
                phonopy_kwargs['fc_name'] = 'force_constants.hdf5'
            else:
                phonopy_kwargs['fc_name'] = 'FORCE_CONSTANTS'
        loaded_data = ForceConstants.from_phonopy(**phonopy_kwargs)

    if loaded_data is None:
        raise ValueError("Could not identify data in Phonopy file.")

    return loaded_data


def load_data_from_file(filename: Union[str, os.PathLike],
                        frequencies_only: bool = False,
                        verbose: bool = False
                        ) -> Union[QpointPhononModes, QpointFrequencies,
                                   ForceConstants]:
    """
    Load phonon mode or force constants data from file

    Parameters
    ----------
    filename
        The file with a path
    frequencies_only
        If true only reads frequencies (not eigenvectors) from the
        file. Only applies if the file is not a force constants
        file.

    Returns
    -------
    file_data
    """
    castep_qpm_suffixes = ('.phonon',)
    castep_fc_suffixes = ('.castep_bin', '.check')
    phonopy_suffixes = ('.hdf5', '.yaml')

    path = pathlib.Path(filename)
    if path.suffix in castep_qpm_suffixes:
        if frequencies_only:
            data = QpointFrequencies.from_castep(path)
        else:
            data = QpointPhononModes.from_castep(path)
    elif path.suffix in castep_fc_suffixes:
        data = ForceConstants.from_castep(path)
    elif path.suffix == '.json':
        data = _load_euphonic_json(path, frequencies_only)
    elif path.suffix in phonopy_suffixes:
        data = _load_phonopy_file(path, frequencies_only)
    else:
        raise ValueError(
            f"File format was not recognised. CASTEP force constants "
            f"data for import should have extension from "
            f"{castep_fc_suffixes}, CASTEP phonon mode data for import "
            f"should have extension '{castep_qpm_suffixes}', data from "
            f"Phonopy should have extension from {phonopy_suffixes}, "
            f"data from Euphonic should have extension '.json'.")
    if verbose:
        print(f'{data.__class__.__name__} data was loaded')
    return data


def get_args(parser: ArgumentParser, params: Optional[List[str]] = None
             ) -> Namespace:
    """
    Get the arguments from the parser. params should only be none when
    running from command line.

    Parameters
    ----------
    parser
        The parser to get the arguments from
    params
        The parameters to get arguments from, if None,
         then parse_args gets them from command line args

    Returns
    -------
    args
        Arguments object for use e.g. args.unit
    """
    if params is None:
        args = parser.parse_args()
    else:
        args = parser.parse_args(params)
    return args


def matplotlib_save_or_show(save_filename: str = None) -> None:
    """
    Save or show the current matplotlib plot.
    Show if save_filename is not None which by default it is.

    Parameters
    ----------
    save_filename
        The file to save the plot in
    """
    import matplotlib.pyplot as plt
    if save_filename is not None:
        plt.savefig(save_filename)
        print(f'Saved plot to {os.path.realpath(save_filename)}')
    else:
        plt.show()


def _get_q_distance(length_unit_string: str, q_distance: float) -> Quantity:
    """
    Parse user arguments to obtain reciprocal-length spacing Quantity
    """
    try:
        length_units = ureg(length_unit_string)
    except UndefinedUnitError:
        raise ValueError("Length unit not known. Euphonic uses Pint for units."
                         " Try 'angstrom' or 'bohr'. Metric prefixes "
                         "are also allowed, e.g 'nm'.")
    recip_length_units = 1 / length_units
    return q_distance * recip_length_units


def _get_energy_bins(
        modes: Union[QpointPhononModes, QpointFrequencies],
        n_ebins: int, emin: Optional[float] = None,
        emax: Optional[float] = None,
        headroom: float = 1.05) -> Quantity:
    """
    Gets recommeded energy bins, in same units as modes.frequencies.
    emin and emax are assumed to be in the same units as
    modes.frequencies, if not provided the min/max values of
    modes.frequencies are used to find the bin limits
    """
    if emin is None:
        # Subtract small amount from min frequency - otherwise due to unit
        # conversions binning of this frequency can vary with different
        # architectures/lib versions, making it difficult to test
        emin_room = 1e-5*ureg('meV').to(modes.frequencies.units).magnitude
        emin = min(np.min(modes.frequencies.magnitude - emin_room), 0.)
    if emax is None:
        emax = np.max(modes.frequencies.magnitude) * headroom
    if emin >= emax:
        raise ValueError("Maximum energy should be greater than minimum. "
                         "Check --e-min and --e-max arguments.")
    ebins = np.linspace(emin, emax, n_ebins) * modes.frequencies.units
    return ebins


def _get_tick_labels(bandpath: dict) -> List[Tuple[int, str]]:
    """Convert x-axis labels from seekpath format to euphonic format

    i.e.::

        ['L', '', '', 'X', '', 'GAMMA']   -->

        [(0, 'L'), (3, 'X'), (5, '$\\Gamma$')]
    """

    label_indices = np.where(bandpath["explicit_kpoints_labels"])[0]
    labels = [bandpath["explicit_kpoints_labels"][i] for i in label_indices]

    for i, label in enumerate(labels):
        if label == 'GAMMA':
            labels[i] = r'$\Gamma$'

    return list(zip(label_indices, labels))


def _get_break_points(bandpath: dict) -> List[int]:
    """Get information about band path labels and break points

    Parameters
    ----------
    bandpath
        Bandpath dictionary from Seekpath

    Returns
    -------
    break_points
        Indices at which the spectrum should be split into subplots
    """
    # Find break points between continuous spectra: wherever there are two
    # adjacent labels
    labels = np.array(bandpath["explicit_kpoints_labels"])

    special_point_bools = np.fromiter(
        map(bool, labels), dtype=bool)

    # [T F F T T F T] -> [F F T T F T] AND [T F F T T F] = [F F F T F F] -> 3,
    adjacent_non_empty_labels = np.logical_and(special_point_bools[:-1],
                                               special_point_bools[1:])

    adjacent_different_labels = (labels[:-1] != labels[1:])

    break_points = np.where(np.logical_and(adjacent_non_empty_labels,
                                           adjacent_different_labels))[0]
    return (break_points + 1).tolist()


def _insert_gamma(bandpath: dict) -> None:
    """Modify seekpath.get_explicit_k_path() results; duplicate Gamma

    This enables LO-TO splitting to be included
    """
    import numpy as np
    gamma_indices = np.where(np.array(bandpath['explicit_kpoints_labels'][1:-1]
                                      ) == 'GAMMA')[0] + 1

    rel_kpts = bandpath['explicit_kpoints_rel'].tolist()
    labels = bandpath['explicit_kpoints_labels']
    for i in reversed(gamma_indices.tolist()):
        rel_kpts.insert(i, [0., 0., 0.])
        labels.insert(i, 'GAMMA')

    bandpath['explicit_kpoints_rel'] = np.array(rel_kpts)
    bandpath['explicit_kpoints_labels'] = labels

    # These unused properties have been invalidated: safer
    # to leave None than incorrect values
    bandpath['explicit_kpoints_abs'] = None
    bandpath['explicit_kpoints_linearcoord'] = None
    bandpath['explicit_segments'] = None


XTickLabels = List[Tuple[int, str]]
SplitArgs = Dict[str, Any]


def _bands_from_force_constants(data: ForceConstants,
                                q_distance: Quantity,
                                insert_gamma: bool = True,
                                frequencies_only: bool = False,
                                **calc_modes_kwargs
                                ) -> Tuple[Union[QpointPhononModes,
                                                 QpointFrequencies],
                                           XTickLabels, SplitArgs]:
    structure = data.crystal.to_spglib_cell()
    bandpath = seekpath.get_explicit_k_path(
        structure,
        reference_distance=q_distance.to('1 / angstrom').magnitude)

    if insert_gamma:
        _insert_gamma(bandpath)

    x_tick_labels = _get_tick_labels(bandpath)
    split_args = {'indices': _get_break_points(bandpath)}

    print(
        "Computing phonon modes: {n_modes} modes across {n_qpts} q-points"
        .format(n_modes=(data.crystal.n_atoms * 3),
                n_qpts=len(bandpath["explicit_kpoints_rel"])))
    qpts = bandpath["explicit_kpoints_rel"]

    if frequencies_only:
        modes = data.calculate_qpoint_frequencies(qpts,
                                                  reduce_qpts=False,
                                                  **calc_modes_kwargs)
    else:
        modes = data.calculate_qpoint_phonon_modes(qpts,
                                                   reduce_qpts=False,
                                                   **calc_modes_kwargs)
    return modes, x_tick_labels, split_args


def _grid_spec_from_args(crystal: Crystal,
                           grid: Optional[Sequence[int]] = None,
                           grid_spacing: Quantity = 0.1 * ureg('1/angstrom')
                           ) -> Tuple[int, int, int]:
    """Get Monkorst-Pack mesh divisions from user arguments"""
    if grid:
        grid_spec = tuple(grid)
    else:
        grid_spec = crystal.get_mp_grid_spec(spacing=grid_spacing)
    return grid_spec


def _get_debye_waller(temperature: Quantity,
                      fc: ForceConstants,
                      grid: Optional[Sequence[int]] = None,
                      grid_spacing: Quantity = 0.1 * ureg('1/angstrom'),
                      **calc_modes_kwargs
                      ) -> DebyeWaller:
    """Generate Debye-Waller data from force constants and grid specification
    """
    mp_grid_spec = _grid_spec_from_args(fc.crystal, grid=grid,
                                        grid_spacing=grid_spacing)
    print("Calculating Debye-Waller factor on {} q-point grid"
          .format(' x '.join(map(str, mp_grid_spec))))
    dw_phonons = fc.calculate_qpoint_phonon_modes(
        euphonic.util.mp_grid(mp_grid_spec),
        **calc_modes_kwargs)
    return dw_phonons.calculate_debye_waller(temperature)


def _get_pdos_weighting(cl_arg_weighting: str) -> Optional[str]:
    """
    Convert CL --weighting to weighting for calculate_pdos
    e.g. --weighting coherent-dos to weighting=coherent
    """
    if cl_arg_weighting == 'dos':
        pdos_weighting = None
    else:
        idx = cl_arg_weighting.rfind('-')
        if idx == -1:
            raise ValueError('Unexpected weighting {cl_arg_weighting}')
        pdos_weighting = cl_arg_weighting[:idx]
    return pdos_weighting


def _arrange_pdos_groups(pdos: Spectrum1DCollection,
                         cl_arg_pdos: Sequence[str]
                         ) -> Union[Spectrum1D, Spectrum1DCollection]:
    """
    Convert PDOS returned by calculate_pdos to PDOS/DOS
    wanted as CL output according to --pdos
    """
    dos = pdos.sum()
    if cl_arg_pdos is not None:
        # Only label total DOS if there are other lines on the plot
        dos.metadata['label'] = 'Total'
        pdos = pdos.group_by('species')
        for line_metadata in pdos.metadata['line_data']:
            line_metadata['label'] = line_metadata['species']
        if len(cl_arg_pdos) > 0:
            pdos = pdos.select(species=cl_arg_pdos)
            dos = pdos
        else:
            dos = Spectrum1DCollection.from_spectra([dos] + [*pdos])
    return dos


def _plot_label_kwargs(args: Namespace, default_xlabel: str = '',
                       default_ylabel: str = '') -> Dict[str, str]:
    """Collect title/label arguments that can be passed to plot_nd
    """
    plot_kwargs = dict(title=args.title,
                       xlabel=default_xlabel,
                       ylabel=default_ylabel)
    if args.ylabel is not None:
        plot_kwargs['ylabel'] = args.ylabel
    if args.xlabel is not None:
        plot_kwargs['xlabel'] = args.xlabel
    return plot_kwargs


def _calc_modes_kwargs(args: Namespace) -> Dict[str, Any]:
    """Collect arguments that can be passed to calculate_qpoint_phonon_modes()
    """
    return dict(asr=args.asr, dipole_parameter=args.dipole_parameter,
                use_c=args.use_c, n_threads=args.n_threads)


def _get_cli_parser(features: Collection[str] = {}
                    ) -> Tuple[ArgumentParser,
                               Dict[str, _ArgumentGroup]]:
    """Instantiate an ArgumentParser with appropriate argument groups

    Parameters
    ----------
    features
        collection (e.g. set, list) of str for known argument groups.
        Known keys: read-fc, read-modes, ins-weighting, pdos-weighting,
        powder, mp-grid, plotting, ebins, adaptive-broadening, q-e,
        map, btol, dipole-parameter-optimisation

    Returns
    -------
    parser
        ArgumentParser with the requested features
    sections
        Dictionary of section labels and their argument groups. This
        allows their help strings to be customised and specialised
        options to be added

    """
    def deprecation_text(deprecated_arg: str, new_arg: str):
        return (f'--{deprecated_arg} is deprecated, '
                f'please use --{new_arg} instead')
    def deprecated_arg(recommended_arg: str):
        class DeprecatedArgAction(Action):
            def __call__(self, parser, args, values, option_string=None):
                # Need to filter to raise warnings from CL tools
                # Warnings not in __main__ are ignored by default
                with warnings.catch_warnings():
                    warnings.filterwarnings(
                        'default', category=DeprecationWarning,
                        module=__name__)
                    warnings.warn(
                        deprecation_text(self.dest, recommended_arg),
                        DeprecationWarning)
                    setattr(args, recommended_arg, values)
        return DeprecatedArgAction

    _pdos_choices = ('coherent-dos', 'incoherent-dos',
                     'coherent-plus-incoherent-dos')
    _ins_choices = ('coherent',)

    parser = ArgumentParser(
        formatter_class=ArgumentDefaultsHelpFormatter)

    # Set up groups; these are only displayed if used, so simplest to
    # instantiate them all now and guarantee consistent order/presence
    # regardless of control logic
    section_defs = [('file', 'File I/O arguments'),
                    ('q', 'q-point sampling arguments'),
                    ('energy', 'energy/frequency arguments'),
                    ('interpolation',
                     'Force constants interpolation arguments'),
                    ('property', 'Property-calculation arguments'),
                    ('plotting', 'Plotting arguments'),
                    ('performance', 'Performance-related arguments')
                    ]

    sections = {label: parser.add_argument_group(doc)
                for label, doc in section_defs}

    if {'read-fc', 'read-modes'}.intersection(features):
        if {'read-fc', 'read-modes'}.issubset(features):
            filename_doc = (
                'Phonon data file. This should contain force constants or '
                'phonon mode data. Force constants formats: .yaml, '
                'force_constants.hdf5 (Phonopy); .castep_bin , .check '
                '(Castep); .json (Euphonic). Phonon mode data formats: '
                '{band,qpoints,mesh}.{hdf5,yaml} (Phonopy); '
                '.phonon (Castep); .json (Euphonic)')
        elif 'read-fc' in features:
            filename_doc = (
                'Phonon data file. This should contain force constants data. '
                'Accepted formats: .yaml, force_constants.hdf5 (Phonopy); '
                '.castep_bin, .check (Castep); .json (Euphonic).')
        else:
            raise ValueError('No band-data-only tools have been defined.')
        sections['file'].add_argument('filename', type=str, help=filename_doc)

    if 'read-fc' in features:
        sections['interpolation'].add_argument(
            '--asr', type=str, nargs='?', default=None, const='reciprocal',
            choices=('reciprocal', 'realspace'),
            help=('Apply an acoustic-sum-rule (ASR) correction to the '
                  'data: "realspace" applies the correction to the force '
                  'constant matrix in real space. "reciprocal" applies '
                  'the correction to the dynamical matrix at each q-point.'))
        if not 'dipole-parameter-optimisation' in features:
            sections['interpolation'].add_argument(
                '--dipole-parameter', type=float, default=1.0,
                dest='dipole_parameter',
                help=('Set the cutoff in real/reciprocal space for the dipole '
                      'Ewald sum; higher values use more reciprocal terms. If '
                      'tuned correctly this can result in performance '
                      'improvements. See euphonic-optimise-dipole-parameter '
                      'program for help on choosing a good DIPOLE_PARAMETER.'))

        use_c = sections['performance'].add_mutually_exclusive_group()
        use_c.add_argument(
            '--use-c', action='store_true', dest='use_c', default=None,
            help=('Force use of compiled C extension when computing '
                  'phonon frequencies/eigenvectors (or raise error).'))
        use_c.add_argument(
            '--disable-c', action='store_false', dest='use_c', default=None,
            help=('Do not attempt to use compiled C extension when computing '
                  'phonon frequencies/eigenvectors.'))
        sections['performance'].add_argument(
            '--n-threads', type=int, default=None, dest='n_threads',
            help=('Number of parallel processes for computing phonon modes. '
                  '(Only applies when using C extension.)'))

    pdos_desc = ('coherent neutron-weighted DOS, incoherent '
                 'neutron-weighted DOS or total (coherent + incoherent) '
                 'neutron-weighted DOS')
    ins_desc = ('coherent inelastic neutron scattering')
    if 'pdos-weighting' in features:
        # Currently do not support multiple PDOS for 2D plots
        if not 'q-e' in features:
            sections['property'].add_argument(
                '--pdos', type=str, action='store', nargs='*',
                help=('Plot PDOS. With --pdos, per-species PDOS will be plotted '
                      'alongside total DOS. A subset of species can also be '
                      'selected by adding more arguments e.g. --pdos Si O'))
        else:
            sections['property'].add_argument(
                '--pdos', type=str, action='store', nargs=1,
                help=('Plot PDOS for a specific species e.g. --pdos Si'))

    if {'pdos-weighting', 'ins-weighting'}.intersection(
            features):
        # We can plot plain DOS with all CL tools
        _weighting_choices = ('dos',)
        desc = 'DOS'
        if 'pdos-weighting' in features:
            _weighting_choices += _pdos_choices
            desc += (', coherent neutron-weighted DOS, incoherent '
                     'neutron-weighted DOS or total (coherent + incoherent) '
                     'neutron-weighted DOS')
        # We may have both 'pdos' and 'ins' so use separate if
        if 'ins-weighting' in features:
            _weighting_choices += _ins_choices
            desc += ', coherent inelastic neutron scattering'
            sections['property'].add_argument(
                '--weighting', '-w', default='dos', choices=_weighting_choices,
                help=(f'Spectral weighting to plot: {desc}'))
        else:
            sections['property'].add_argument(
                '--weighting', '-w', default='dos', choices=_weighting_choices,
                help=f'Type of DOS to plot: {desc}')

    if 'ins-weighting' in features:
        sections['property'].add_argument(
            '--temperature', type=float, default=None,
            help=('Temperature in K; enable Debye-Waller factor calculation. '
                  '(Only applicable when --weighting=coherent).'))

    # 'ins-weighting' implies 'mp-grid' as well because mesh is needed for DW
    # factor
    if {'ins-weighting', 'mp-grid'}.intersection(features):
        grid_spec = sections['q'].add_mutually_exclusive_group()

        grid_spec.add_argument(
            '--grid', type=int, nargs=3, default=None,
            help=('Defines a Monkhorst-Pack grid.'))
        grid_spec.add_argument(
            '--grid-spacing', type=float, default=0.1, dest='grid_spacing',
            help=('q-point spacing of Monkhorst-Pack grid.'))

    if 'powder' in features:
        _sampling_choices = {'golden', 'sphere-projected-grid',
                             'spherical-polar-grid',
                             'spherical-polar-improved',
                             'random-sphere'}
        npts_group = sections['q'].add_mutually_exclusive_group()
        npts_group.add_argument('--npts', '-n', type=int, default=1000,
                                help=('Number of samples at each |q| sphere.'))
        npts_group.add_argument(
            '--npts-density', type=int, default=None, dest='npts_density',
            help=('NPTS specified as the number of points at surface of '
                  '1/LENGTH_UNIT-radius sphere; otherwise scaled to equivalent'
                  ' area density at sphere surface.'))

        sections['q'].add_argument(
            '--npts-min', type=int, default=100, dest='npts_min',
            help=('Minimum number of samples per sphere. This ensures adequate'
                  ' sampling at small q when using --npts-density.'))
        sections['q'].add_argument(
            '--npts-max', type=int, default=10000, dest='npts_max',
            help=('Maximum number of samples per sphere. This avoids '
                  'diminishing returns at large q when using --npts-density.'))
        sections['q'].add_argument(
            '--sampling', type=str, default='golden',
            choices=_sampling_choices,
            help=('Spherical sampling scheme; "golden" is generally '
                  'recommended uniform quasirandom sampling.'))
        sections['q'].add_argument(
            '--jitter', action='store_true',
            help=('Apply additional jitter to sample positions in angular '
                  'direction. Recommended for sampling methods other than '
                  '"golden" and "random-sphere".'))

    if 'plotting' in features:
        sections['file'].add_argument(
            '--save-json', dest='save_json', default=None,
            help='Save spectrum to a .json file with this name')
        section = sections['plotting']
        section.add_argument(
            '-s', '--save-to', dest='save_to', default=None,
            help='Save resulting plot to a file with this name')
        section.add_argument('--title', type=str, default='',
                             help='Plot title')
        section.add_argument('--x-label', '--xlabel', type=str, default=None,
                             dest='xlabel', help='Plot x-axis label')
        section.add_argument('--y-label', '--ylabel', type=str, default=None,
                             dest='ylabel', help='Plot y-axis label')
        section.add_argument('--style', type=str, nargs='+',
                             help='Matplotlib styles (name or file)')
        section.add_argument('--no-base-style', action='store_true',
                             dest='no_base_style',
                             help=('Remove all default formatting before '
                                   'applying other style options.'))
        section.add_argument('--font', type=str, default=None,
                             help=('Select text font. (This has to be a name '
                                   'known to Matplotlib. font-family will be '
                                   'set to sans-serif; it doesn\'t matter if)'
                                   'the font is actually sans-serif.'))
        section.add_argument('--font-size', '--fontsize', type=float,
                             default=None, dest='fontsize',
                             help='Set base font size in pt.')
        section.add_argument('--fig-size', '--figsize', type=float, nargs=2,
                             default=None, dest='figsize',
                             help='Figure canvas size in FIGSIZE-UNITS')
        section.add_argument('--fig-size-unit', '--figsize-unit', type=str,
                             default='cm', dest='figsize_unit',
                             help='Unit of length for --figsize')

    if ('plotting' in features) and not ('map' in features):
        section = sections['plotting']
        section.add_argument('--line-width', '--linewidth', type=float,
                             default=None, dest='linewidth',
                             help='Set line width in pt.')

    if {'ebins', 'q-e'}.intersection(features):
        section = sections['energy']
        section.add_argument('--e-min', type=float, default=None, dest='e_min',
                             help='Energy range minimum in ENERGY_UNIT')
        section.add_argument('--e-max', type=float, default=None, dest='e_max',
                             help='Energy range maximum in ENERGY_UNIT')
        section.add_argument('--energy-unit', '-u', dest='energy_unit',
                             type=str, default='meV', help='Energy units')

    if {'ebins', 'adaptive-broadening'}.intersection(features):
        if 'ebins' in features:
            section = sections['energy']
            section.add_argument('--ebins', type=int, default=200,
                                 help='Number of energy bins')
            if 'adaptive-broadening' in features:
                section.add_argument(
                    '--adaptive', action='store_true',
                    help=('Use adaptive broadening on the energy axis to '
                          'broaden based on phonon mode widths, rather than '
                          'using fixed width broadening'))
                section.add_argument(
                    '--adaptive-method', type=str, default='reference',
                    dest='adaptive_method', choices=('reference', 'fast'),
                    help='The adaptive broadening method')
                section.add_argument(
                    '--adaptive-error', type=float, default=0.01,
                    dest='adaptive_error',
                    help=('Maximum absolute error '
                          'for gaussian approximations '
                          'when using the fast adaptive '
                          'broadening method'))
                eb_help = (
                    'If using fixed width broadening, the FWHM of broadening '
                    'on energy axis in ENERGY_UNIT (no broadening if '
                    'unspecified). If using adaptive broadening, this is a '
                    'scale factor multiplying the mode widths (no scale '
                    'factor applied if unspecified).')
            else:
                eb_help = ('The FWHM of broadening on energy axis in '
                           'ENERGY_UNIT (no broadening if unspecified).')
            section.add_argument(
                '--energy-broadening', '--eb', type=float, default=None,
                dest='energy_broadening',
                help=eb_help)
            section.add_argument(
                '--shape', type=str, nargs='?', default='gauss',
                choices=('gauss', 'lorentz'),
                help='The broadening shape')
        else:
            ValueError('"adaptive-broadening" cannot be applied without '
                       '"ebins"')

    if {'q-e', 'mp-grid'}.intersection(features):
        sections['q'].add_argument(
            '--length-unit', type=str, default='angstrom', dest='length_unit',
            help=('Length units; these will be inverted to obtain '
                  'units of distance between q-points (e.g. "bohr"'
                  ' for bohr^-1).'))

    if 'q-e' in features:
        sections['q'].add_argument(
            '--q-spacing', type=float, dest='q_spacing', default=0.025,
            help=('Target distance between q-point samples in 1/LENGTH_UNIT'))

    if {'q-e', 'map'}.issubset(features):
        sections['q'].add_argument(
            '--q-broadening', '--qb', type=float, default=None,
            dest='q_broadening',
            help=('FWHM of broadening on q axis in 1/LENGTH_UNIT '
                  '(no broadening if unspecified).'))

    if 'map' in features:
        sections['plotting'].add_argument(
            '--v-min', '--vmin', type=float, default=None, dest='vmin',
            help='Minimum of data range for colormap.')
        sections['plotting'].add_argument(
            '--v-max', '--vmax', type=float, default=None, dest='vmax',
            help='Maximum of data range for colormap.')
        sections['plotting'].add_argument(
            '--c-map', '--cmap', type=str, default=None, dest='cmap',
            help='Matplotlib colormap')

    if 'btol' in features:
        sections['q'].add_argument(
            '--btol', default=10.0, type=float,
            help=('Distance threshold used for automatically splitting '
                  'discontinuous segments of reciprocal space onto separate '
                  'subplots. This is specified as a multiple of the median '
                  'distance between q-points.'))

    if 'dipole-parameter-optimisation' in features:
        parser.add_argument(
            '-n',
            default=500,
            type=int,
            help=('The number of times to loop over q-points. A higher '
                  'value will get a more reliable timing, but will take '
                  'longer')
        )
        parser.add_argument(
            '--dipole-parameter-min', '--min',
            default=0.25,
            type=float,
            help='The minimum value of dipole_parameter to test'
        )
        parser.add_argument(
            '--dipole-parameter-max', '--max',
            default=1.5,
            type=float,
            help='The maximum value of dipole_parameter to test'
        )
        parser.add_argument(
            '--dipole-parameter-step', '--step',
            default=0.25,
            type=float,
            help='The difference between each dipole_parameter to test'
        )

    return parser, sections


MplStyle = Union[str, Dict[str, str]]


def _compose_style(
        *, user_args: Namespace, base: Optional[List[MplStyle]]
        ) -> List[MplStyle]:
    """Combine user-specified style options with default stylesheets

    Args:
        user_args: from _get_cli_parser().parse_args()
        base: Euphonic default styles for this plot

    N.B. matplotlib applies styles from left to right, so the right-most
    elements of the list take the highest priority. This function builds a
    list in the order:

    [base style(s), user style(s), CLI arguments]
    """

    if user_args.no_base_style or base is None:
        style = []
    else:
        style = base

    if user_args.style:
        style += user_args.style

    # Explicit args take priority over any other
    explicit_args = {}
    for user_arg, mpl_property in {'cmap': 'image.cmap',
                                   'fontsize': 'font.size',
                                   'font': 'font.sans-serif',
                                   'linewidth': 'lines.linewidth',
                                   'figsize': 'figure.figsize'}.items():
        if getattr(user_args, user_arg, None):
            explicit_args.update({mpl_property: getattr(user_args, user_arg)})

    if 'font.sans-serif' in explicit_args:
        explicit_args.update({'font.family': 'sans-serif'})

    if 'figure.figsize' in explicit_args:
        dimensioned_figsize = [dim * ureg(user_args.figsize_unit)
                               for dim in explicit_args['figure.figsize']]
        explicit_args['figure.figsize'] = [dim.to('inches').magnitude
                                           for dim in dimensioned_figsize]

    style.append(explicit_args)
    return style



"""Matplotlib stylesheets for plot styling"""
from pkg_resources import resource_filename

base_style = resource_filename("euphonic.styles", "base.mplstyle")
intensity_widget_style = resource_filename("euphonic.styles",
                                           "intensity_widget.mplstyle")



import inspect
from typing import TypeVar, Dict, Any, Type

from euphonic.validate import _check_constructor_inputs, _check_unit_conversion
from euphonic.io import (_obj_to_json_file, _obj_from_json_file,
                         _obj_to_dict, _process_dict)
from euphonic import ureg, Quantity, Crystal


class DebyeWaller:
    """
    Stores the (n_ions, 3, 3) anisotropic Debye-Waller exponent
    W_ab to be used in structure factor calculations

    Attributes
    ----------
    crystal
        Lattice and atom information
    debye_waller
        Shape (n_ions, 3, 3) float Quantity. The anisotropic
        Debye-Waller exponent W_ab for each atom, where the
        Debye-Waller factor is exp(-W_ab*Q_a*Q_b) where a,b run over
        the 3 Cartesian directions
    temperature
        Scalar float Quantity. The temperature the Debye-Waller
        exponent was calculated at
    """
    T = TypeVar('T', bound='DebyeWaller')

    def __init__(self, crystal: Crystal, debye_waller: Quantity,
                 temperature: Quantity) -> None:
        """
        Parameters
        ----------
        crystal
            Lattice and atom information
        debye_waller
            Shape (n_atoms, 3, 3) Quantity. The anisotropic
            Debye-Waller exponent W_ab for each atom, where the
            Debye-Waller factor is exp(-W_ab*Q_a*Q_b) where a,b run
            over the 3 Cartesian directions
        temperature
            Scalar float Quantity. The temperature the Debye-Waller
            exponent was calculated at
        """
        _check_constructor_inputs([crystal], [Crystal], [()], ['crystal'])
        n_atoms = crystal.n_atoms
        _check_constructor_inputs(
            [debye_waller, temperature],
            [Quantity, Quantity],
            [(n_atoms, 3, 3), ()],
            inspect.getfullargspec(self.__init__)[0][2:])
        self.crystal = crystal
        self._debye_waller = debye_waller.to(ureg.bohr**2).magnitude
        self._temperature = temperature.to(ureg.K).magnitude

        self.debye_waller_unit = str(debye_waller.units)
        self.temperature_unit = str(temperature.units)

    @property
    def debye_waller(self) -> Quantity:
        return self._debye_waller*ureg('bohr**2').to(
            self.debye_waller_unit)

    @debye_waller.setter
    def debye_waller(self, value: Quantity) -> None:
        self.debye_waller_unit = str(value.units)
        self._debye_waller = value.to('bohr**2').magnitude

    @property
    def temperature(self) -> Quantity:
        # See https://pint.readthedocs.io/en/latest/nonmult.html
        return Quantity(self._temperature, ureg('K')).to(self.temperature_unit)

    @temperature.setter
    def temperature(self, value: Quantity) -> None:
        self.temperature_unit = str(value.units)
        self._temperature = value.to('K').magnitude

    def __setattr__(self, name: str, value: Any) -> None:
        _check_unit_conversion(self, name, value,
                               ['debye_waller_unit', 'temperature_unit'])
        super(DebyeWaller, self).__setattr__(name, value)

    def to_dict(self) -> Dict[str, Any]:
        """
        Convert to a dictionary. See DebyeWaller.from_dict for details
        on keys/values

        Returns
        -------
        dict
        """
        dout = _obj_to_dict(self, ['crystal', 'debye_waller', 'temperature'])
        return dout

    def to_json_file(self, filename: str) -> None:
        """
        Write to a JSON file. JSON fields are equivalent to
        DebyeWaller.from_dict keys

        Parameters
        ----------
        filename
            Name of the JSON file to write to
        """
        _obj_to_json_file(self, filename)

    @classmethod
    def from_dict(cls: Type[T], d: Dict[str, Any]) -> T:
        """
        Convert a dictionary to a DebyeWaller object

        Parameters
        ----------
        d
            A dictionary with the following keys/values:

            - 'crystal': dict, see Crystal.from_dict
            - 'debye_waller': (n_atoms, 3, 3) float ndarray
            - 'debye_waller_unit': str
            - 'temperature': float
            - 'temperature_unit': str

        Returns
        -------
        DebyeWaller
        """
        crystal = Crystal.from_dict(d['crystal'])
        d = _process_dict(d, quantities=['debye_waller', 'temperature'],
                          optional=['temperature'])
        return cls(crystal, d['debye_waller'], d['temperature'])

    @classmethod
    def from_json_file(cls: Type[T], filename: str) -> T:
        """
        Read from a JSON file. See DebyeWaller.from_dict for required
        fields

        Parameters
        ----------
        filename
            The file to read from

        Returns
        -------
        DebyeWaller
        """
        return _obj_from_json_file(cls, filename)
